import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from enum import Enum
import qis as qis

import quant_strats.local_path as lp

output_path = lp.OUTPUT_PATH
local_path = lp.LOCAL_PATH
# quant_strats
from quant_strats import (compute_strategy_portfolio_data_with_costs,
                          compute_multi_strategy_data_from_blocks,
                          compute_marginal_perfs_for_strategy)
from quant_strats.data.universes.futures.bbg_futures import Universes
from quant_strats.data.set.assets_bbg import AssetsBBG
from quant_strats.research_strategies.futures.cross_trend.csft_strategies import (CSTF_RB_TRACKER,
                                                                                  CSTF_EXACT_TRACKER,
                                                                                  CSTF_RB_AC_TRACKER)
from quant_strats.research_strategies.futures.trackers.run_trackers_backtest import (CSTF_GOLDMAN_TRACKER_AC,
                                                                                     GOLDMAN_UNIVERSE)

strategy_universe = Universes.BBG_FUTURES_INVESTABLE.load_universe_data(local_path=local_path)
prices = strategy_universe.get_prices(freq='B')[['UXY1 Comdty', 'US1 Comdty', 'WN1 Comdty']].ffill().dropna()
prices = prices[prices.columns[::-1]]
qis.plot_prices_with_dd(prices=prices)

portfolio_data = qis.backtest_model_portfolio(prices=prices,
                                              weights=np.array([-1.0, 2.0, -1.0]),
                                              rebalance_freq='ME',
                                              ticker='Butterly'
                                              )

time_period = qis.TimePeriod('31Dec2015', None)
figs = qis.generate_strategy_factsheet(portfolio_data=portfolio_data,
                                       benchmark_prices=prices.iloc[:, -1],
                                       is_grouped=False,
                                       add_current_position_var_risk_sheet=False,
                                       add_weights_turnover_sheet=False,
                                       add_grouped_exposures=False,
                                       add_grouped_cum_pnl=False,
                                       time_period=time_period,
                                       **qis.fetch_default_report_kwargs(time_period=time_period,
                                                                         add_rates_data=False))

qis.save_figs_to_pdf(figs=figs, file_name=f"butterfly", orientation='landscape',
                     local_path=output_path)

plt.show()


"""
core implementation of baktest report focus on strategy and comparison vs benchmark stategy
"""

# packages
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from dataclasses import dataclass
from typing import List, Optional, Tuple, Union

# qis
import qis as qis
from qis import PerfParams, PerfStat, RegimeData, BenchmarkReturnsQuantileRegimeSpecs, TimePeriod, RollingPerfStat
import qis.utils.struct_ops as sop
import qis.utils.df_groups as dfg
import qis.perfstats.returns as ret
import qis.perfstats.perf_stats as rpt
import qis.perfstats.regime_classifier as rcl

# plots
import qis.plots.time_series as pts
import qis.plots.derived.prices as ppd
import qis.plots.derived.returns_heatmap as rhe
import qis.plots.derived.perf_table as ppt
import qis.plots.derived.returns_scatter as prs
import qis.plots.derived.drawdowns as cdr
from qis.portfolio.portfolio_data import PortfolioData, AttributionMetric


PERF_PARAMS = PerfParams(freq='W-WED')
REGIME_PARAMS = BenchmarkReturnsQuantileRegimeSpecs(freq='ME')


@dataclass
class MultiPortfolioData:
    """
    data structure to unify multi portfolio reporting
    """
    portfolio_datas: List[PortfolioData]
    benchmark_prices: Union[pd.DataFrame, pd.Series] = None

    def __post_init__(self):
        # default frequency is freq of backtests, can be non for strats at different freqs
        self.set_navs(freq=None)
        if self.benchmark_prices is not None and isinstance(self.benchmark_prices, pd.Series):
            self.benchmark_prices = self.benchmark_prices.to_frame()

    def set_navs(self, freq: Optional[str] = None) -> None:
        navs = []
        for portfolio in self.portfolio_datas:
            navs.append(portfolio.get_portfolio_nav())
        self.navs = pd.concat(navs, axis=1)

        if freq is not None:
            self.navs = self.navs.asfreq(freq=freq, method='ffill')

        if self.benchmark_prices is not None:
            self.benchmark_prices = self.benchmark_prices.reindex(index=self.navs.index, method='ffill')

    def set_benchmark_prices(self, benchmark_prices: pd.DataFrame) -> None:
        # can pass benchmark prices here
        if isinstance(benchmark_prices, pd.Series):
            benchmark_prices = benchmark_prices.to_frame()
        else:
            benchmark_prices = benchmark_prices
        self.benchmark_prices = benchmark_prices.reindex(index=self.navs.index, method='ffill')

    """
    data get methods
    """
    def get_navs(self,
                 add_benchmarks_to_navs: bool = False,
                 benchmark: str = None,
                 time_period: TimePeriod = None
                 ) -> pd.DataFrame:
        """
        get portfolio navs
        """
        navs = self.navs
        if benchmark is not None:
            navs = pd.concat([self.benchmark_prices[benchmark].reindex(index=navs.index).ffill(), navs], axis=1)
        elif add_benchmarks_to_navs:
            navs = pd.concat([navs, self.benchmark_prices.reindex(index=navs.index).ffill()], axis=1).ffill()
            
        if time_period is not None:
            navs = time_period.locate(navs)
        return navs

    def get_benchmark_price(self,
                            benchmark: str,
                            time_period: TimePeriod = None
                            ) -> pd.Series:
        price = self.benchmark_prices[benchmark]
        if time_period is not None:
            price = time_period.locate(price)
        return price

    def get_group_navs(self,
                       portfolio_idx: int = 0,
                       benchmark: str = None,
                       time_period: TimePeriod = None,
                       is_add_group_total: bool = False
                       ) -> pd.DataFrame:
        prices = self.portfolio_datas[portfolio_idx].get_group_navs(time_period=time_period, is_add_group_total=is_add_group_total)
        if benchmark is not None:
            benchmark_price = self.get_benchmark_price(benchmark=self.benchmark_prices.columns[0],
                                                       time_period=time_period)
            prices = pd.concat([prices, benchmark_price], axis=1)
        return prices

    """
    plot methods
    """
    def add_regime_shadows(self,
                           ax: plt.Subplot,
                           regime_benchmark: str,
                           index: pd.Index = None,
                           regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS
                           ) -> None:
        """
        add regime shadows using regime_benchmark
        """
        pivot_prices = self.benchmark_prices[regime_benchmark]
        if index is not None:
            pivot_prices = pivot_prices.reindex(index=index, method='ffill')
        qis.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)

    def plot_nav(self,
                 time_period: TimePeriod = None,
                 regime_benchmark: str = None,
                 perf_params: PerfParams = PERF_PARAMS,
                 regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                 add_benchmarks_to_navs: bool = False,
                 ax: plt.Subplot = None,
                 **kwargs) -> None:

        if ax is None:
            fig, ax = plt.subplots()

        prices = self.get_navs(time_period=time_period, add_benchmarks_to_navs=add_benchmarks_to_navs)
        ppd.plot_prices(prices=prices,
                        perf_params=perf_params,
                        ax=ax,
                        **kwargs)
        if regime_benchmark is not None:
            self.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark, index=prices.index, regime_params=regime_params)

    def plot_rolling_perf(self,
                          rolling_perf_stat: RollingPerfStat = RollingPerfStat.SHARPE,
                          regime_benchmark: str = None,
                          time_period: TimePeriod = None,
                          sharpe_rolling_window: int = 260,
                          freq_sharpe: Optional[str] = None,
                          sharpe_title: Optional[str] = None,
                          legend_stats: pts.LegendStats = pts.LegendStats.AVG_LAST,
                          regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                          add_benchmarks_to_navs: bool = False,
                          ax: plt.Subplot = None,
                          **kwargs
                          ) -> plt.Figure:

        # do not use start end dates here so the sharpe will be continuous with different time_period
        if ax is None:
            fig, ax = plt.subplots()

        prices = self.get_navs(time_period=time_period, add_benchmarks_to_navs=add_benchmarks_to_navs)

        kwargs = qis.update_kwargs(kwargs, dict(roll_freq=freq_sharpe))
        fig = ppd.plot_rolling_perf_stat(prices=prices,
                                         rolling_perf_stat=rolling_perf_stat,
                                         time_period=time_period,
                                         roll_periods=sharpe_rolling_window,
                                         legend_stats=legend_stats,
                                         trend_line=None,  # qis.TrendLine.ZERO_SHADOWS,
                                         ax=ax,
                                         **kwargs)

        if regime_benchmark is not None:
            self.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark, index=prices.index,
                                    regime_params=regime_params)
        return fig
    
    def plot_periodic_returns(self,
                              time_period: TimePeriod = None,
                              heatmap_freq: str = 'YE',
                              date_format: str = '%Y',
                              transpose: bool = True,
                              add_benchmarks_to_navs: bool = False,
                              title: str = None,
                              ax: plt.Subplot = None,
                              **kwargs
                              ) -> None:
        prices = self.get_navs(time_period=time_period, add_benchmarks_to_navs=add_benchmarks_to_navs)
        rhe.plot_periodic_returns_table(prices=prices,
                                        freq=heatmap_freq,
                                        ax=ax,
                                        title=title,
                                        date_format=date_format,
                                        transpose=transpose,
                                        **kwargs)

    def plot_performance_bars(self,
                              time_period: TimePeriod = None,
                              perf_column: PerfStat = PerfStat.SHARPE_RF0,
                              perf_params: PerfParams = PERF_PARAMS,
                              add_benchmarks_to_navs: bool = False,
                              title: str = None,
                              ax: plt.Subplot = None,
                              **kwargs
                              ) -> None:
        prices = self.get_navs(time_period=time_period, add_benchmarks_to_navs=add_benchmarks_to_navs)
        ppt.plot_ra_perf_bars(prices=prices,
                              perf_column=perf_column,
                              perf_params=perf_params,
                              title=title or f"{perf_column.to_str()}: {qis.get_time_period(prices).to_str()}",
                              ax=ax,
                              **kwargs)

    def plot_corr_table(self,
                        time_period: TimePeriod = None,
                        freq: str = 'W-WED',
                        add_benchmarks_to_navs: bool = False,
                        ax: plt.Subplot = None,
                        **kwargs) -> None:
        prices = self.get_navs(time_period=time_period, add_benchmarks_to_navs=add_benchmarks_to_navs)
        if len(prices.columns) > 1:
            qis.plot_returns_corr_table(prices=prices,
                                        x_rotation=90,
                                        freq=freq,
                                        title=f'Correlation of {freq} returns',
                                        ax=ax,
                                        **kwargs)

    def plot_drawdowns(self,
                       time_period: TimePeriod = None,
                       regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                       regime_benchmark: str = None,
                       dd_legend_type: qis.DdLegendType = qis.DdLegendType.SIMPLE,
                       add_benchmarks_to_navs: bool = False,
                       ax: plt.Subplot = None,
                       **kwargs) -> None:
        prices = self.get_navs(time_period=time_period, add_benchmarks_to_navs=add_benchmarks_to_navs)
        cdr.plot_rolling_drawdowns(prices=prices, dd_legend_type=dd_legend_type, ax=ax, **kwargs)
        if regime_benchmark is not None:
            self.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark, index=prices.index, regime_params=regime_params)

    def plot_rolling_time_under_water(self,
                                      time_period: TimePeriod = None,
                                      regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                                      regime_benchmark: str = None,
                                      dd_legend_type: qis.DdLegendType = qis.DdLegendType.SIMPLE,
                                      add_benchmarks_to_navs: bool = False,
                                      ax: plt.Subplot = None,
                                      **kwargs) -> None:
        prices = self.get_navs(time_period=time_period, add_benchmarks_to_navs=add_benchmarks_to_navs)
        cdr.plot_rolling_time_under_water(prices=prices, dd_legend_type=dd_legend_type, ax=ax, **kwargs)
        if regime_benchmark is not None:
            self.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark, index=prices.index, regime_params=regime_params)

    def get_ra_perf_table(self,
                          benchmark: str = None,
                          time_period: TimePeriod = None,
                          drop_benchmark: bool = False,
                          is_convert_to_str: bool = True,
                          perf_params: PerfParams = PERF_PARAMS,
                          perf_columns: List[PerfStat] = rpt.BENCHMARK_TABLE_COLUMNS,
                          **kwargs
                          ) -> pd.DataFrame:
        if benchmark is None:
            benchmark = self.benchmark_prices.columns[0]
        prices = self.get_navs(benchmark=benchmark, time_period=time_period)
        ra_perf_table = ppt.get_ra_perf_benchmark_columns(prices=prices,
                                                          benchmark=benchmark,
                                                          drop_benchmark=drop_benchmark,
                                                          is_convert_to_str=is_convert_to_str,
                                                          perf_params=perf_params,
                                                          perf_columns=perf_columns,
                                                          **kwargs)
        return ra_perf_table

    def plot_ra_perf_table(self,
                           benchmark: str = None,
                           time_period: TimePeriod = None,
                           drop_benchmark: bool = False,
                           perf_params: PerfParams = PERF_PARAMS,
                           perf_columns: List[PerfStat] = rpt.BENCHMARK_TABLE_COLUMNS,
                           ax: plt.Subplot = None,
                           **kwargs
                           ) -> None:
        if benchmark is None:
            benchmark = self.benchmark_prices.columns[0]
        prices = self.get_navs(benchmark=benchmark, time_period=time_period)
        ra_perf_title = f"RA performance table for {perf_params.freq_vol}-freq returns with beta to {benchmark}: {qis.get_time_period(prices).to_str()}"
        ppt.plot_ra_perf_table_benchmark(prices=prices,
                                         benchmark=benchmark,
                                         perf_params=perf_params,
                                         perf_columns=perf_columns,
                                         drop_benchmark=drop_benchmark,
                                         title=ra_perf_title,
                                         rotation_for_columns_headers=0,
                                         ax=ax,
                                         **kwargs)

    def plot_ac_ra_perf_table(self,
                              benchmark_price: pd.Series,
                              add_benchmarks_to_navs: bool = False,
                              time_period: TimePeriod = None,
                              perf_params: PerfParams = PERF_PARAMS,
                              perf_columns: List[PerfStat] = rpt.BENCHMARK_TABLE_COLUMNS,
                              is_grouped: bool = True,
                              ax: plt.Subplot = None,
                              **kwargs) -> None:

        if add_benchmarks_to_navs:
            prices = self.get_navs(time_period=time_period, add_benchmarks_to_navs=add_benchmarks_to_navs)
            drop_benchmark = False
            rows_edge_lines = [len(self.portfolio_datas)]
        else:
            if is_grouped:  # otherwise tables look too bad
                add_ac = True
                drop_benchmark = True
            else:
                add_ac = False
                drop_benchmark = False

            strategy_prices = []
            ac_prices = []
            rows_edge_lines = [len(self.portfolio_datas)]
            for portfolio in self.portfolio_datas:
                portfolio_name = str(portfolio.nav.name)
                navs_ = portfolio.get_portfolio_nav(time_period=time_period)  # navs include costs while group navs are cost free
                ac_prices_ = portfolio.get_group_navs(time_period=time_period, is_add_group_total=False)
                strategy_prices.append(navs_)
                if add_ac:
                    ac_prices_.columns = [f"{portfolio_name}-{x}" for x in ac_prices_.columns]
                    ac_prices.append(ac_prices_)
                    rows_edge_lines.append(sum(rows_edge_lines)+len(ac_prices_.columns))
            strategy_prices = pd.concat(strategy_prices, axis=1)

            benchmark_price = benchmark_price.reindex(index=strategy_prices.index, method='ffill')
            if add_ac:  # otherwise tables look too bad
                ac_prices = pd.concat(ac_prices, axis=1)
                prices = pd.concat([benchmark_price, strategy_prices, ac_prices],axis=1)
            else:
                prices = pd.concat([strategy_prices, benchmark_price], axis=1)

        ra_perf_title = f"RA performance table for {perf_params.freq_vol}-freq returns with beta to {benchmark_price.name}: {qis.get_time_period(prices).to_str()}"
        ppt.plot_ra_perf_table_benchmark(prices=prices,
                                         benchmark=str(benchmark_price.name),
                                         perf_params=perf_params,
                                         perf_columns=perf_columns,
                                         drop_benchmark=drop_benchmark,
                                         rows_edge_lines=rows_edge_lines,
                                         title=ra_perf_title,
                                         rotation_for_columns_headers=0,
                                         row_height=0.5,
                                         ax=ax,
                                         **kwargs)

    def plot_exposures(self,
                       benchmark: str = None,
                       regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                       time_period: TimePeriod = None,
                       var_format: str = '{:.0%}',
                       ax: plt.Subplot = None,
                       **kwargs
                       ) -> None:
        exposures = []
        for portfolio in self.portfolio_datas:
            exposures.append(portfolio.get_weights(time_period=time_period).sum(axis=1).rename(portfolio.nav.name))
        exposures = pd.concat(exposures, axis=1)
        pts.plot_time_series(df=exposures,
                             var_format=var_format,
                             legend_stats=pts.LegendStats.AVG_NONNAN_LAST,
                             title='Portfolio net exposures',
                             ax=ax,
                             **kwargs)
        if benchmark is not None:
            self.add_regime_shadows(ax=ax, regime_benchmark=benchmark, index=exposures.index, regime_params=regime_params)

    def plot_instrument_pnl_diff(self,
                                 portfolio_idx1: int = 0,
                                 portfolio_idx2: int = 1,
                                 is_grouped: bool = True,
                                 benchmark: str = None,
                                 regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                                 time_period: TimePeriod = None,
                                 var_format: str = '{:.0%}',
                                 ax: plt.Subplot = None,
                                 **kwargs) -> None:
        pnl_inst1 = self.portfolio_datas[portfolio_idx1].instrument_pnl
        pnl_inst2 = self.portfolio_datas[portfolio_idx2].instrument_pnl
        df1_, df2_ = pnl_inst1.align(other=pnl_inst2, join='outer', axis=None)
        df1_, df2_ = df1_.fillna(0.0), df2_.fillna(0.0)
        diff = df1_.subtract(df2_)

        if is_grouped:
            diff = dfg.agg_df_by_groups_ax1(diff,
                                            group_data=self.portfolio_datas[portfolio_idx1].group_data,
                                            agg_func=np.nansum,
                                            total_column=f"{self.portfolio_datas[portfolio_idx1].nav.name}-{self.portfolio_datas[portfolio_idx2].nav.name}",
                                            group_order=self.portfolio_datas[portfolio_idx1].group_order)
        if time_period is not None:
            diff = time_period.locate(diff)
        diff = diff.cumsum(axis=0)

        pts.plot_time_series(df=diff,
                             var_format=var_format,
                             legend_stats=pts.LegendStats.LAST_NONNAN,
                             title=f"Cumulative p&l diff {self.portfolio_datas[portfolio_idx1].nav.name}-{self.portfolio_datas[portfolio_idx2].nav.name}",
                             ax=ax,
                             **sop.update_kwargs(kwargs, dict(legend_loc='lower left')))
        if benchmark is not None:
            self.add_regime_shadows(ax=ax, regime_benchmark=benchmark, index=diff.index, regime_params=regime_params)

    def plot_exposures_diff(self,
                            portfolio_idx1: int = 0,
                            portfolio_idx2: int = 1,
                            benchmark: str = None,
                            regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                            time_period: TimePeriod = None,
                            var_format: str = '{:.0%}',
                            ax: plt.Subplot = None,
                            **kwargs) -> None:
        exposures1 = self.portfolio_datas[portfolio_idx1].get_weights(is_grouped=True, time_period=time_period, add_total=False)
        exposures2 = self.portfolio_datas[portfolio_idx2].get_weights(is_grouped=True, time_period=time_period, add_total=False)
        diff = exposures1.subtract(exposures2)
        pts.plot_time_series(df=diff,
                             var_format=var_format,
                             legend_stats=pts.LegendStats.AVG_NONNAN_LAST,
                             title=f"Net exposure diff {self.portfolio_datas[portfolio_idx1].nav.name}-{self.portfolio_datas[portfolio_idx2].nav.name}",
                             ax=ax,
                             **kwargs)
        if benchmark is not None:
            self.add_regime_shadows(ax=ax, regime_benchmark=benchmark, index=diff.index, regime_params=regime_params)

    def plot_turnover(self,
                      benchmark: str = None,
                      time_period: TimePeriod = None,
                      regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                      turnover_rolling_period: Optional[int] = 260,
                      freq_turnover: Optional[str] = 'B',
                      var_format: str = '{:.0%}',
                      ax: plt.Subplot = None,
                      **kwargs) -> None:

        turnover = []
        for portfolio in self.portfolio_datas:
            turnover.append(portfolio.get_turnover(roll_period=turnover_rolling_period, freq=freq_turnover, is_agg=True).rename(portfolio.nav.name))
        turnover = pd.concat(turnover, axis=1)
        if time_period is not None:
            turnover = time_period.locate(turnover)
        freq = pd.infer_freq(turnover.index)
        turnover_title = f"{turnover_rolling_period}-period rolling {freq}-freq Turnover"
        pts.plot_time_series(df=turnover,
                             var_format=var_format,
                             y_limits=(0.0, None),
                             legend_stats=pts.LegendStats.AVG_NONNAN_LAST,
                             title=turnover_title,
                             ax=ax,
                             **kwargs)
        if benchmark is not None:
            self.add_regime_shadows(ax=ax, regime_benchmark=benchmark, index=turnover.index, regime_params=regime_params)

    def plot_costs(self,
                   cost_rolling_period: Optional[int] = 260,
                   freq_cost: Optional[str] = 'B',
                   cost_title: Optional[str] = None,
                   benchmark: str = None,
                   time_period: TimePeriod = None,
                   regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                   var_format: str = '{:.2%}',
                   is_norm_costs: bool = True,
                   ax: plt.Subplot = None,
                   **kwargs) -> None:
        costs = []
        for portfolio in self.portfolio_datas:
            costs.append(portfolio.get_costs(roll_period=cost_rolling_period, freq=freq_cost, is_agg=True, is_norm_costs=is_norm_costs).rename(portfolio.nav.name))
        costs = pd.concat(costs, axis=1)
        if time_period is not None:
            costs = time_period.locate(costs)
        freq = pd.infer_freq(costs.index)
        cost_title = cost_title or f"{cost_rolling_period}-period rolling {freq}-freq Costs %"
        pts.plot_time_series(df=costs,
                             var_format=var_format,
                             y_limits=(0.0, None),
                             legend_stats=pts.LegendStats.AVG_NONNAN_LAST,
                             title=cost_title,
                             ax=ax,
                             **kwargs)
        if benchmark is not None:
            self.add_regime_shadows(ax=ax, regime_benchmark=benchmark, index=costs.index, regime_params=regime_params)

    def plot_factor_betas(self,
                          benchmark_prices: pd.DataFrame,
                          regime_benchmark: str = None,
                          time_period: TimePeriod = None,
                          regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                          freq_beta: str = 'B',
                          factor_beta_span: int = 260,
                          var_format: str = '{:,.2f}',
                          axs: List[plt.Subplot] = None,
                          **kwargs
                          ) -> None:
        """
        plot benchmarks betas by factor exposures
        """
        factor_exposures = {factor: [] for factor in benchmark_prices.columns}
        for portfolio in self.portfolio_datas:
            factor_exposure = portfolio.compute_portfolio_benchmark_betas(benchmark_prices=benchmark_prices,
                                                                          freq_beta=freq_beta,
                                                                          factor_beta_span=factor_beta_span,
                                                                          time_period=time_period)
            for factor in factor_exposure.columns:
                factor_exposures[factor].append(factor_exposure[factor].rename(portfolio.nav.name))

        if axs is None:
            fig, axs = plt.subplots(len(benchmark_prices.columns), 1, figsize=(12, 12), tight_layout=True)

        for idx, factor in enumerate(benchmark_prices.columns):
            factor_exposure = pd.concat(factor_exposures[factor], axis=1)
            factor_beta_title = f"{factor_beta_span}-span rolling Beta of {freq_beta}-freq returns to {factor}"
            pts.plot_time_series(df=factor_exposure,
                                 var_format=var_format,
                                 legend_stats=pts.LegendStats.AVG_NONNAN_LAST,
                                 title=f"{factor_beta_title}",
                                 ax=axs[idx],
                                 **kwargs)
            if regime_benchmark is not None:
                self.add_regime_shadows(ax=axs[idx],
                                        regime_benchmark=regime_benchmark,
                                        index=factor_exposure.index,
                                        regime_params=regime_params)

    def plot_nav_with_dd(self,
                         time_period: TimePeriod = None,
                         perf_params: PerfParams = PERF_PARAMS,
                         axs: List[plt.Subplot] = None,
                         **kwargs
                         ) -> None:
        prices = self.get_navs(time_period=time_period)
        if self.benchmark_prices is not None:
            regime_benchmark_str = self.benchmark_prices.columns[0]
        else:
            regime_benchmark_str = None
        ppd.plot_prices_with_dd(prices=prices,
                                perf_params=perf_params,
                                regime_benchmark_str=regime_benchmark_str,
                                axs=axs,
                                **kwargs)

    def plot_returns_scatter(self,
                             benchmark: str,
                             time_period: TimePeriod = None,
                             freq: str = 'QE',
                             order: int = 2,
                             ax: plt.Subplot = None,
                             **kwargs
                             ) -> None:
        prices = self.get_navs(benchmark=benchmark, time_period=time_period)
        local_kwargs = sop.update_kwargs(kwargs=kwargs,
                                         new_kwargs={'weight': 'bold',
                                                     'x_rotation': 0,
                                                     'first_color_fixed': False,
                                                     'ci': None,
                                                     'markersize': 6})
        prs.plot_returns_scatter(prices=prices,
                                 benchmark=benchmark,
                                 freq=freq,
                                 order=order,
                                 title=f"Scatterplot of {freq}-freq returns vs {benchmark}",
                                 ax=ax,
                                 **local_kwargs)

    def plot_performance_attribution(self,
                                     portfolio_ids: List[int] = (0, ),
                                     time_period: TimePeriod = None,
                                     attribution_metric: AttributionMetric = AttributionMetric.PNL,
                                     ax: plt.Subplot = None,
                                     **kwargs
                                     ) -> None:
        datas = []
        for portfolio_id in portfolio_ids:
            datas.append(self.portfolio_datas[portfolio_id].get_performance_attribution_data(attribution_metric=attribution_metric,
                                                                                             time_period=time_period))
        data = pd.concat(datas, axis=1)
        data = data.sort_values(data.columns[0], ascending=False)
        kwargs = sop.update_kwargs(kwargs=kwargs,
                                         new_kwargs={'ncols': len(data.columns),
                                                     'legend_loc': 'upper center',
                                                     'bbox_to_anchor': (0.5, 1.05),
                                                     'x_rotation': 90})
        data = data.replace({0.0: np.nan}).dropna()
        qis.plot_bars(df=data,
                      skip_y_axis=True,
                      title=f"{attribution_metric.title}",
                      stacked=False,
                      ax=ax,
                      **kwargs)

    def plot_performance_periodic_table(self,
                                        portfolio_id: int = 0,
                                        time_period: TimePeriod = None,
                                        freq: str = 'YE',
                                        ax: plt.Subplot = None,
                                        **kwargs
                                        ) -> None:

        inst_returns = self.portfolio_datas[portfolio_id].get_attribution_table_by_instrument(time_period=time_period)
        inst_navs = ret.returns_to_nav(returns=inst_returns, init_period=None)
        strategy_nav = self.portfolio_datas[portfolio_id].get_portfolio_nav(time_period=time_period)
        prices = pd.concat([inst_navs, strategy_nav], axis=1).dropna()
        rhe.plot_periodic_returns_table(prices=prices,
                                        title=f"{strategy_nav.name} Attribution by Instrument",
                                        freq=freq,
                                        ax=ax,
                                        **kwargs)

    def plot_regime_data(self,
                         benchmark: str,
                         add_benchmarks_to_navs: bool = False,
                         is_grouped: bool = False,
                         portfolio_idx: int = 0,
                         regime_data_to_plot: RegimeData = RegimeData.REGIME_SHARPE,
                         time_period: TimePeriod = None,
                         var_format: Optional[str] = None,
                         is_conditional_sharpe: bool = True,
                         is_use_vbar: bool = False,
                         perf_params: PerfParams = PERF_PARAMS,
                         regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                         title: str = None,
                         legend_loc: Optional[str] = 'upper center',
                         ax: plt.Subplot = None,
                         **kwargs
                         ) -> None:
        if is_grouped:
            prices = self.get_group_navs(portfolio_idx=portfolio_idx, benchmark=benchmark, time_period=time_period)
        else:
            prices = self.get_navs(benchmark=benchmark, add_benchmarks_to_navs=add_benchmarks_to_navs, time_period=time_period)
        if var_format is None:
            if regime_data_to_plot == RegimeData.REGIME_SHARPE:
                var_format = '{:.2f}'
            else:
                var_format = '{:.2%}'

        title = title or f"Sharpe ratio split to {str(benchmark)} Bear/Normal/Bull {regime_params.freq}-freq regimes"
        regime_classifier = rcl.BenchmarkReturnsQuantilesRegime(regime_params=regime_params)
        qis.plot_regime_data(regime_classifier=regime_classifier,
                             prices=prices,
                             benchmark=benchmark,
                             is_conditional_sharpe=is_conditional_sharpe,
                             regime_data_to_plot=regime_data_to_plot,
                             var_format=var_format,
                             is_use_vbar=is_use_vbar,
                             regime_params=regime_params,
                             legend_loc=legend_loc,
                             perf_params=perf_params,
                             title=title,
                             ax=ax,
                             **kwargs)

    def compute_brinson_attribution(self,
                                    strategy_idx: int = 0,
                                    benchmark_idx: int = 1,
                                    freq: Optional[str] = None,
                                    total_column: str = 'Total Sum',
                                    time_period: TimePeriod = None,
                                    is_exclude_interaction_term: bool = True
                                    ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:

        strategy_pnl = self.portfolio_datas[strategy_idx].get_attribution_table_by_instrument(time_period=time_period, freq=freq)
        strategy_weights = self.portfolio_datas[strategy_idx].get_weights(time_period=time_period, freq=freq, is_input_weights=False)

        benchmark_pnl = self.portfolio_datas[benchmark_idx].get_attribution_table_by_instrument(time_period=time_period, freq=freq)
        benchmark_weights = self.portfolio_datas[benchmark_idx].get_weights(time_period=time_period, freq=freq, is_input_weights=False)

        asset_class_data = self.portfolio_datas[strategy_idx].group_data
        group_order = self.portfolio_datas[strategy_idx].group_order

        totals_table, active_total, grouped_allocation_return, grouped_selection_return, grouped_interaction_return = \
            qis.compute_brinson_attribution_table(benchmark_pnl=benchmark_pnl,
                                                  strategy_pnl=strategy_pnl,
                                                  strategy_weights=strategy_weights,
                                                  benchmark_weights=benchmark_weights,
                                                  asset_class_data=asset_class_data,
                                                  group_order=group_order,
                                                  total_column=total_column,
                                                  is_exclude_interaction_term=is_exclude_interaction_term,
                                                  strategy_name=self.portfolio_datas[strategy_idx].ticker or 'Strategy',
                                                  benchmark_name=self.portfolio_datas[benchmark_idx].ticker or 'Benchmark')
        return totals_table, active_total, grouped_allocation_return, grouped_selection_return, grouped_interaction_return

    def plot_brinson_attribution(self,
                                 strategy_idx: int = 0,
                                 benchmark_idx: int = 1,
                                 freq: Optional[str] = None,
                                 axs: List[plt.Subplot] = (None, None, None, None, None),
                                 total_column: str = 'Total Sum',
                                 time_period: TimePeriod = None,
                                 is_exclude_interaction_term: bool = True,
                                 **kwargs
                                 ) -> Tuple[plt.Figure, plt.Figure, plt.Figure, plt.Figure, plt.Figure]:

        totals_table, active_total, grouped_allocation_return, grouped_selection_return, grouped_interaction_return = \
            self.compute_brinson_attribution(strategy_idx=strategy_idx,
                                             benchmark_idx=benchmark_idx,
                                             freq=freq,
                                             total_column=total_column,
                                             time_period=time_period,
                                             is_exclude_interaction_term=is_exclude_interaction_term)

        fig_table, fig_active_total, fig_ts_alloc, fig_ts_sel, fig_ts_inters = qis.plot_brinson_attribution_table(
            totals_table=totals_table,
            active_total=active_total,
            grouped_allocation_return=grouped_allocation_return,
            grouped_selection_return=grouped_selection_return,
            grouped_interaction_return=grouped_interaction_return,
            total_column=total_column,
            is_exclude_interaction_term=is_exclude_interaction_term,
            axs=axs,
            **kwargs)

        return fig_table, fig_active_total, fig_ts_alloc, fig_ts_sel, fig_ts_inters

    def plot_weights_boxplot(self,
                             strategy_idx: int = 0,
                             benchmark_idx: int = 1,
                             freq: Optional[str] = 'ME',
                             time_period: TimePeriod = None,
                             is_grouped: bool = False,
                             ax: plt.Subplot = None,
                             **kwargs
                             ) -> None:
        strategy_weights = self.portfolio_datas[strategy_idx].get_weights(time_period=time_period, freq=freq,
                                                                          is_grouped=is_grouped, is_input_weights=False)
        benchmark_weights = self.portfolio_datas[benchmark_idx].get_weights(time_period=time_period, freq=freq,
                                                                            is_grouped=is_grouped, is_input_weights=False)
        dfs = {self.portfolio_datas[strategy_idx].ticker: strategy_weights,
               self.portfolio_datas[benchmark_idx].ticker: benchmark_weights}

        local_kwargs = qis.update_kwargs(kwargs, dict(ncols=2, legend_loc='upper center', showmedians=True,
                                                      yvar_format='{:.1%}'))
        qis.df_dict_boxplot_by_columns(dfs=dfs,
                                       hue_var_name='groups' if is_grouped else 'instruments',
                                       y_var_name='weights',
                                       ylabel='weights',
                                       ax=ax,
                                       **local_kwargs)

    def plot_group_exposures_and_pnl(self,
                                     regime_benchmark: str = None,
                                     time_period: TimePeriod = None,
                                     regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                                     total_name: str = 'Total',
                                     exposures_freq: Optional[str] = 'W-WED',
                                     figsize: Tuple[float, float] = (8.3, 11.7),  # A4 for portrait
                                     **kwargs
                                     ) -> List[plt.Figure]:
        """
        report total exposures by portfolio
        """
        # aling to the first
        group_data = self.portfolio_datas[0].group_data
        group_order = self.portfolio_datas[0].group_order
        grouped_exposures_aggs = {}  # dict[portfolio, Dict[group, pd.Dataframe]]
        grouped_pnls_aggs = {}
        for portfolio_data in self.portfolio_datas:
            grouped_exposures_agg, grouped_exposures_by_inst = portfolio_data.get_grouped_long_short_exposures(time_period=time_period,
                                                                                                               exposures_freq=exposures_freq,
                                                                                                               total_name=total_name)
            grouped_pnls_agg, grouped_pnls_by_inst = portfolio_data.get_grouped_cum_pnls(time_period=time_period,
                                                                                         total_name=total_name)

            grouped_exposures_aggs[portfolio_data.ticker] = grouped_exposures_agg
            grouped_pnls_aggs[portfolio_data.ticker] = grouped_pnls_agg

        group_exposures_by_portfolio = {group: {} for group in group_order}
        for group in group_data:
            for ticker, df in grouped_exposures_aggs.items():
                group_exposures_by_portfolio[group].update({ticker: df[group][total_name]})

        group_pnl_by_portfolio = {group: {} for group in group_order}
        for group in group_data:
            for ticker, df in grouped_pnls_aggs.items():
                group_pnl_by_portfolio[group].update({ticker: df[group][total_name]})

        for group in group_order:
            group_exposures_by_portfolio[group] = pd.DataFrame.from_dict(group_exposures_by_portfolio[group], orient='columns')
            group_pnl_by_portfolio[group] = pd.DataFrame.from_dict(group_pnl_by_portfolio[group], orient='columns')

        figs = []
        for group, exposures_agg in group_exposures_by_portfolio.items():
            fig, axs = plt.subplots(2, 1, figsize=figsize, tight_layout=True)
            qis.set_suptitle(fig, f"Total Exposures and P&L by {group}")
            figs.append(fig)
            qis.plot_time_series(df=exposures_agg,
                                 var_format='{:,.0%}',
                                 legend_stats=qis.LegendStats.AVG_MIN_MAX_LAST,
                                 title=f"Exposure by {group}",
                                 ax=axs[0],
                                 **kwargs)
            qis.plot_time_series(df=group_pnl_by_portfolio[group],
                                 var_format='{:,.0%}',
                                 legend_stats=qis.LegendStats.LAST,
                                 title=f"Cumulative P&L by {group}",
                                 ax=axs[1],
                                 **kwargs)
            if regime_benchmark is not None:
                for ax in axs:
                    self.add_regime_shadows(ax=ax,
                                            regime_benchmark=regime_benchmark,
                                            index=exposures_agg.index,
                                            regime_params=regime_params)

        return figs

    def compute_tracking_error_table(self,
                                     strategy_idx: int = 0,
                                     benchmark_idx: int = 1,
                                     freq: Optional[str] = 'B',
                                     time_period: TimePeriod = None,
                                     af: float = 260,
                                     is_norm_costs: bool = True,
                                     **kwargs
                                     ) -> pd.DataFrame:

        strategy_pnl = self.portfolio_datas[strategy_idx].get_attribution_table_by_instrument(time_period=time_period, freq=freq)
        benchmark_pnl = self.portfolio_datas[benchmark_idx].get_attribution_table_by_instrument(time_period=time_period, freq=freq)

        tickers_union = qis.merge_lists_unique(list1=strategy_pnl.columns.to_list(), list2=benchmark_pnl.columns.to_list())
        strategy_pnl = strategy_pnl.reindex(columns=tickers_union)
        benchmark_pnl = benchmark_pnl.reindex(columns=tickers_union).reindex(index=strategy_pnl.index)
        pnl_diff = strategy_pnl.subtract(benchmark_pnl)

        strategy_weights = self.portfolio_datas[strategy_idx].get_weights(time_period=time_period, freq=None, is_input_weights=True)
        strategy_turnover = self.portfolio_datas[strategy_idx].get_turnover(time_period=time_period, freq=None, roll_period=None, add_total=False)
        strategy_cost = self.portfolio_datas[strategy_idx].get_costs(time_period=time_period, freq=freq, roll_period=None,
                                                                     add_total=False, is_norm_costs=is_norm_costs)
        strategy_ticker = self.portfolio_datas[strategy_idx].ticker

        benchmark_weights = self.portfolio_datas[benchmark_idx].get_weights(time_period=time_period, freq=None, is_input_weights=True)
        benchmark_turnover = self.portfolio_datas[benchmark_idx].get_turnover(time_period=time_period, freq=None, roll_period=None, add_total=False)
        benchmark_cost = self.portfolio_datas[benchmark_idx].get_costs(time_period=time_period, freq=freq, roll_period=None,
                                                                       add_total=False, is_norm_costs=is_norm_costs)
        benchmark_ticker = self.portfolio_datas[benchmark_idx].ticker

        # compute stats
        total_strategy_perf = strategy_pnl.cumsum(0).iloc[-1, :].rename(f"{strategy_ticker} total perf")
        total_benchmark_perf = benchmark_pnl.cumsum(0).iloc[-1, :].rename(f"{benchmark_ticker} total perf")
        total_diff = total_strategy_perf.subtract(total_benchmark_perf).rename(f"{strategy_ticker}-{benchmark_ticker} total perf")

        tre = pd.Series(np.nanmean(pnl_diff, axis=0) / np.nanstd(pnl_diff, axis=0), index=tickers_union, name='TRE')

        tre_table = pd.concat([total_diff, tre,
                               total_strategy_perf, total_benchmark_perf,
                               af*strategy_turnover.mean(0).rename(f"{strategy_ticker} an turnover"),
                               af*benchmark_turnover.mean(0).rename(f"{benchmark_ticker} an turnover"),
                               af*strategy_cost.mean(0).rename(f"{strategy_ticker} an cost"),
                               af*benchmark_cost.mean(0).rename(f"{benchmark_ticker} an cost"),
                               ], axis=1)

        return tre_table

    def get_aligned_weights(self,
                            strategy_idx: int = 0,
                            benchmark_idx: int = 1,
                            freq: Optional[str] = 'B',
                            time_period: TimePeriod = None,
                            is_grouped: bool = False,
                            **kwargs
                            ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        strategy_weights = self.portfolio_datas[strategy_idx].get_weights(time_period=time_period, freq=None, is_input_weights=True, is_grouped=is_grouped)
        benchmark_weights = self.portfolio_datas[benchmark_idx].get_weights(time_period=time_period, freq=None, is_input_weights=True, is_grouped=is_grouped)
        tickers_union = qis.merge_lists_unique(list1=strategy_weights.columns.to_list(), list2=benchmark_weights.columns.to_list())
        if is_grouped and self.portfolio_datas[benchmark_idx].group_order is not None: # replace with ac order of benchmark
            tickers_union = self.portfolio_datas[benchmark_idx].group_order
        strategy_weights = strategy_weights.reindex(columns=tickers_union)
        benchmark_weights = benchmark_weights.reindex(columns=tickers_union)
        return strategy_weights, benchmark_weights

    def get_aligned_turnover(self,
                             strategy_idx: int = 0,
                             benchmark_idx: int = 1,
                             turnover_rolling_period: Optional[int] = 260,
                             freq_turnover: Optional[str] = 'B',
                             time_period: TimePeriod = None,
                             is_grouped: bool = False,
                             **kwargs
                             ) -> Tuple[pd.DataFrame, pd.DataFrame]:

        strategy_turnover = self.portfolio_datas[strategy_idx].get_turnover(time_period=time_period, freq=freq_turnover,
                                                                            roll_period=turnover_rolling_period,
                                                                            add_total=False, is_grouped=is_grouped)
        benchmark_turnover = self.portfolio_datas[benchmark_idx].get_turnover(time_period=time_period, freq=freq_turnover,
                                                                              roll_period=turnover_rolling_period,
                                                                              add_total=False, is_grouped=is_grouped)
        tickers_union = qis.merge_lists_unique(list1=strategy_turnover.columns.to_list(), list2=benchmark_turnover.columns.to_list())
        if is_grouped and self.portfolio_datas[benchmark_idx].group_order is not None: # replace with ac order of benchmark
            tickers_union = self.portfolio_datas[benchmark_idx].group_order
        strategy_turnover = strategy_turnover.reindex(columns=tickers_union)
        benchmark_turnover = benchmark_turnover.reindex(columns=tickers_union)
        return strategy_turnover, benchmark_turnover



"""
implement winsorizing of time series data using ewm
1. compute mean_t and vol_t
2. select x% of outliers defined by normalized score (x_t-mean_t) / vol_t
3. replace or trim outliers as specified
"""
# packages
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from numba import njit
from enum import Enum
from typing import Union, NamedTuple, Optional, Tuple
from qis.models.linear.ewm import compute_ewm, compute_ewm_vol


class ReplacementType(Enum):
    EWMA_MEAN = 1
    NAN = 2
    QUANTILES = 3


class OutlierPolicy(NamedTuple):
    """
    specify filtering policy params
    """
    abs_ceil: Optional[float] = None  # remove all above
    abs_floor: Optional[float] = None  # remove all below
    std_abs_ceil: Optional[float] = None  # > 0
    std_abs_floor: Optional[float] = None  # < 0
    std_ewm_ceil: Optional[float] = None  # >0
    std_ewm_floor: Optional[float] = None  # <0
    ewm_lambda: Union[float, np.ndarray] = 0.94
    is_log_transform: bool = False
    nan_replacement_type: ReplacementType = ReplacementType.NAN


class OutlierPolicyTypes(OutlierPolicy, Enum):
    """
    defined policy type
    """
    HARD_CEIL_POLICY = OutlierPolicy(abs_floor=0.0001,
                                     std_abs_ceil=10.0)

    RANGE_CEIL_POLICY = OutlierPolicy(abs_floor=0.0001,
                                      std_abs_ceil=10.0)

    SOFT_RANGE_CEIL_POLICY = OutlierPolicy(abs_floor=1e-8,
                                           std_ewm_ceil=10.0,
                                           std_ewm_floor=None,
                                           std_abs_ceil=10.0)

    SOFT_POSITIVE_LOG_POLICY = OutlierPolicy(abs_floor=1e-8,
                                             std_ewm_ceil=10.0,
                                             std_ewm_floor=None,
                                             is_log_transform=True)
    NONE = None


def filter_outliers(data: Union[pd.DataFrame, pd.Series, np.ndarray],
                    outlier_policy: OutlierPolicy
                    ) -> Union[pd.DataFrame, pd.Series, np.ndarray]:

    np.seterr(invalid='ignore')  # off warnings

    if isinstance(data, pd.DataFrame) or isinstance(data, pd.Series):
        orig_data = data.to_numpy()
    elif isinstance(data, np.ndarray):
        orig_data = data
    else:
        raise TypeError('filter_outliers: unsupported data type')

    clean_data = orig_data.copy()
    #  keep track of nans - nans will be put back to output data
    non_nan_cond = np.isfinite(orig_data)

    # imnitial replacement is using nans
    nan_replacement = np.full_like(orig_data, np.nan, dtype=np.float64)

    # remove absolute outliers
    if outlier_policy.abs_ceil is not None:
        clean_data = np.where(np.greater(clean_data, outlier_policy.abs_ceil, where=non_nan_cond),
                              nan_replacement, clean_data)
    if outlier_policy.abs_floor is not None:
        clean_data = np.where(np.less(clean_data, outlier_policy.abs_floor, where=non_nan_cond),
                              nan_replacement, clean_data)

    # now apply log transform
    if outlier_policy.is_log_transform:
        if outlier_policy.abs_floor is None:
            raise TypeError('is_log_transform must be applied with abs_floor > 0')
        log_cond = np.greater(clean_data, 0.0, where=non_nan_cond)
        clean_data = np.log(clean_data, where=log_cond)
    else:
        log_cond = None

    # remove relative outliers to in-sample std
    if outlier_policy.std_abs_ceil is not None or outlier_policy.std_abs_floor is not None:
        nan_mean = np.nanmean(clean_data, axis=0)
        nan_std = np.nanstd(clean_data, axis=0)

        if outlier_policy.std_abs_ceil is not None:
            ceil = np.add(nan_mean, nan_std*outlier_policy.std_abs_ceil, where=non_nan_cond)
            clean_data = np.where(np.greater(clean_data, ceil, where=non_nan_cond), nan_replacement, clean_data)

        if outlier_policy.std_abs_floor is not None:
            floor = np.add(nan_mean, nan_std*outlier_policy.std_abs_floor, where=non_nan_cond)
            clean_data = np.where(np.less(clean_data, floor, where=non_nan_cond), nan_replacement, clean_data)

    # now rolling ewm outliers
    if outlier_policy.std_ewm_ceil is not None or outlier_policy.std_ewm_floor is not None:

        ewm_mean, score = compute_ewm_score(data=clean_data, ewm_lambda=outlier_policy.ewm_lambda)
        if outlier_policy.std_ewm_ceil is not None:
            clean_data = np.where(np.greater(score, outlier_policy.std_ewm_ceil, where=non_nan_cond),
                                  nan_replacement, clean_data)

        if outlier_policy.std_ewm_floor is not None:
            clean_data = np.where(np.less(score, outlier_policy.std_ewm_floor, where=non_nan_cond),
                                  nan_replacement, clean_data)
    if outlier_policy.is_log_transform:
        clean_data = np.exp(clean_data, where=log_cond)

    # implemented replacement type is EWMA mean
    if outlier_policy.nan_replacement_type == ReplacementType.EWMA_MEAN:
        ewm_mean, _ = compute_ewm_score(data=clean_data, ewm_lambda=outlier_policy.ewm_lambda)
        filtered_data = np.where(np.isfinite(clean_data), clean_data, ewm_mean)
    else:
        filtered_data = np.where(non_nan_cond, clean_data, nan_replacement)

    if isinstance(data, pd.DataFrame):
        filtered_data = pd.DataFrame(data=filtered_data, columns=data.columns, index=data.index)
    elif isinstance(data, pd.Series):
        filtered_data = pd.Series(data=filtered_data, name=data.name, index=data.index)

    return filtered_data


def ewm_insample_winsorising(data: Union[pd.DataFrame, pd.Series, np.ndarray],
                             ewm_lambda: Union[float, np.ndarray] = 0.94,
                             quantile_cut: float = 0.025,
                             nan_replacement_type: ReplacementType = ReplacementType.EWMA_MEAN
                             ) -> Union[pd.DataFrame, pd.Series, np.ndarray]:

    if isinstance(data, pd.DataFrame) or isinstance(data, pd.Series):
        np_data = data.to_numpy()
    elif isinstance(data, np.ndarray):
        np_data = data.copy()
    else:
        raise TypeError('ewm_winsorising: unsupported data type')

    # 1 compute ewm score
    ewm_mean, score = compute_ewm_score(data=np_data, ewm_lambda=ewm_lambda)

    lower_quantile = np.quantile(score, quantile_cut, axis=0)
    upper_quantile = np.quantile(score, 1.0-quantile_cut, axis=0)
    #print(f"lower_quantile={lower_quantile}, upper_quantile={upper_quantile}")

    if nan_replacement_type == ReplacementType.EWMA_MEAN:
        replacement_cond = np.logical_or(score < lower_quantile, score > upper_quantile)
        winsor_data = np.where(replacement_cond, ewm_mean, np_data)

    elif nan_replacement_type == ReplacementType.NAN:
        replacement_cond = np.logical_or(score < lower_quantile, score > upper_quantile)
        winsor_data = np.where(replacement_cond, np.full_like(np_data, np.nan), np_data)

    elif nan_replacement_type == ReplacementType.QUANTILES:
        winsor_data = np.where(score < lower_quantile, np.quantile(np_data, quantile_cut, axis=0), np_data)
        winsor_data = np.where(score > upper_quantile, np.quantile(np_data, 1.0-quantile_cut, axis=0), winsor_data)
    else:
        raise TypeError('replacement_type not implemented')

    if isinstance(data, pd.DataFrame):
        winsor_data = pd.DataFrame(data=winsor_data, columns=data.columns, index=data.index)
    elif isinstance(data, pd.Series):
        winsor_data = pd.Series(data=winsor_data, name=data.name, index=data.index)

    return winsor_data


def compute_ewm_score(data: np.ndarray,
                      ewm_lambda: Union[float, np.ndarray] = 0.94,
                      is_clip: bool = True,
                      clip_quantile: float = 0.16
                      ) -> (np.ndarray, np.ndarray):

    ewm_mean = compute_ewm(data=data, ewm_lambda=ewm_lambda)
    ewm_vol = compute_ewm_vol(data=data, ewm_lambda=ewm_lambda)
    if is_clip:  # remove small values below 1 _ std quantile
        ewm_vol = np.clip(a=ewm_vol, a_min=np.nanquantile(ewm_vol, clip_quantile), a_max=None)
    non_nan_cond = np.isfinite(data)
    score = np.divide(np.subtract(data, ewm_mean), ewm_vol, where=non_nan_cond)
    return ewm_mean, score


# @njit
def ewm_winsdor_markovian_score(a: np.ndarray,
                                init_value: Union[float, np.ndarray],
                                init_var: Union[float, np.ndarray] = None,
                                score_threshold: float = 5.0,
                                span: Union[int, np.ndarray] = 31,
                                ewm_lambda: Union[float, np.ndarray] = None,
                                is_start_from_first_nonan: bool = True
                                ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:

    """
    use ewma score to filter out outliers in non-anticipating markovian way
    data: numpy with dimension = t*n

    score_t is defined as non-anticipating:
    score_t = (x[t]-ewm[t-1] / np.sqrt(ewm2[t-1])
    outlier x[t] is defined when:
    np.abs(score_t) > threshold

    if x[t] is outlier, it is ignored for ewm:
    ewm[t] = ewm[t-1]
    ewm2[t] = ewm2[t-1]
    else ewm is computed using recursion:
    ewm[t] = (1-lambda) * x[t] + lambda*ewm[t-1]
    ewm2[t] = (1-lambda) * (x[t]-ewm[t])^2 + lambda*ewm2[t-1]

    if x[t] is nan:
    ewm[t] = ewm[t-1]
    ewm2[t] = ewm2[t-1]

    assumption is that no np.nan value is returned from the function

    ewm_lambda: float or ndarray of dimension n
    init_value: initial value of dimension n
    start_from_first_nonan: start filling nans only from the first non-nan in underlying data: recomended because
                            it avoids backfilling of init_value
    """
    if span is not None:
        ewm_lambda = 1.0 - 2.0 / (span + 1.0)
    ewm_lambda_1 = 1.0 - ewm_lambda

    is_1d = (a.ndim == 1)  # or a.shape[1] == 1)

    # initialize all
    ewm = np.full_like(a, fill_value=np.nan, dtype=np.double)
    ewm2 = np.full_like(a, fill_value=np.nan, dtype=np.double)
    score = np.full_like(a, fill_value=np.nan, dtype=np.double)
    clean_a = np.full_like(a, fill_value=np.nan, dtype=np.double)

    if init_var is None:
        if is_1d:
            init_var = 0.1
        else:
            init_var = 0.1*np.ones(a.shape[1])

    if is_start_from_first_nonan:
        if is_1d:  # cannot use np.where
            last_ewm = init_value if np.isfinite(a[0]) else np.nan
        else:
            last_ewm = np.where(np.isfinite(a[0]), init_value, np.nan)
    else:
        last_ewm = init_value

    last_ewm2 = np.maximum(last_ewm * last_ewm, init_var)

    ewm[0] = last_ewm
    ewm2[0] = last_ewm2
    score[0] = 0.0

    # recurse from 1
    for t in np.arange(1, a.shape[0]):
        a_t = a[t]

        if is_start_from_first_nonan:
            # detect starting nonnans for when last ewma was np.nan and a_t is finite
            if is_1d:  # cannot use np.where
                if np.isfinite(last_ewm) == False and np.isfinite(a_t) == True:  # trick: if last_ewm is nan
                    last_ewm = init_value
                    last_ewm2 = np.maximum(init_value*init_value, init_var)
            else:
                new_nonnans = np.logical_and(np.isfinite(last_ewm)==False, np.isfinite(a_t)==True)
                if np.any(new_nonnans):
                    last_ewm = np.where(new_nonnans, init_value, last_ewm)
                    last_ewm2 = np.where(new_nonnans, np.maximum(init_value*init_value, init_var), last_ewm2)

        # fill nan-values
        current_ewm_ = ewm_lambda * last_ewm + ewm_lambda_1 * a_t
        current_ewm2_ = ewm_lambda * last_ewm2 + ewm_lambda_1 * np.square(a_t-current_ewm_)

        # score_t = np.divide(a_t - last_ewm, np.sqrt(last_ewm2), where=np.greater(last_ewm2, 0.0))
        score_t = np.where(np.greater(last_ewm2, 0.0), (a_t - last_ewm)/np.sqrt(last_ewm2), np.nan)
        is_outlier = np.abs(score_t) >= score_threshold

        if is_1d:   # np.where cannot be used
            if is_outlier:
                current_ewm = last_ewm
                current_ewm2 = last_ewm2
                clean_a_ = clean_a[t-1]
            else:
                current_ewm = current_ewm_
                current_ewm2 = current_ewm2_
                clean_a_ = a_t
        else:
            current_ewm = np.where(is_outlier, current_ewm_, last_ewm)
            current_ewm2 = np.where(is_outlier, current_ewm2_, last_ewm2)
            clean_a_ = np.where(is_outlier, clean_a[t-1], a_t)

        ewm[t] = last_ewm = current_ewm
        ewm2[t] = last_ewm2 = current_ewm2
        score[t] = score_t
        clean_a[t] = clean_a_

    return clean_a, ewm, ewm2, score


class UnitTests(Enum):
    TEST1 = 1
    MARKOVIAN_WINDSOR = 2


def run_unit_test(unit_test: UnitTests):

    np.random.seed(2) #freeze seed

    import qis as qis

    dates = pd.date_range(start='12/31/2018', end='12/31/2019', freq='B')
    n = 2

    data = pd.DataFrame(data=np.random.standard_t(df=2, size=(len(dates), n)),
                        index=dates,
                        columns=['x'+str(m+1) for m in range(n)])
    # data = pd.Series(data=data_np, index=dates, name='data')

    if unit_test == UnitTests.TEST1:
        winsor_data = ewm_insample_winsorising(data=data,
                                               ewm_lambda=0.94,
                                               nan_replacement_type=ReplacementType.EWMA_MEAN,
                                               quantile_cut=0.05)

        winsor_data.columns = [x + ' winsor' for x in data.columns]

        plot_data = pd.concat([data, winsor_data], axis=1)
        title = 'Data Winsor'

        qis.plot_time_series(df=plot_data,
                             title=title,
                             legend_loc='upper left',
                             legend_stats=qis.LegendStats.AVG,
                             last_label=qis.LastLabel.AVERAGE_VALUE,
                             trend_line=qis.TrendLine.AVERAGE,
                             var_format='{:.2f}')

    elif unit_test == UnitTests.MARKOVIAN_WINDSOR:
        clean_x, ewm, ewm2, score = ewm_winsdor_markovian_score(a=data.to_numpy(), span=7,
                                                                init_value=0.0,
                                                                init_var=0.1*np.ones(len(data.columns)))
        clean_x = pd.DataFrame(clean_x, index=dates, columns=[x + ' winsor' for x in data.columns])
        plot_data = pd.concat([data, clean_x], axis=1)
        qis.plot_time_series(df=plot_data,
                             title='Windsor',
                             legend_loc='upper left',
                             legend_stats=qis.LegendStats.AVG_STD_LAST,
                             var_format='{:.2f}')

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.MARKOVIAN_WINDSOR

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
create total return blended portfolios based on weights
"""

# packages
import numpy as np
import pandas as pd
from numba import njit
from enum import Enum
from typing import Union, Dict, Tuple, List, Optional

# qis
import qis.utils as qu
from qis.portfolio.portfolio_data import PortfolioData


def backtest_model_portfolio(prices: pd.DataFrame,
                             weights: Union[Dict[str, float], List[float], np.ndarray, pd.DataFrame, pd.Series],
                             rebalance_freq: Optional[str] = 'QE',
                             initial_nav: float = 100,
                             funding_rate: pd.Series = None,  # on positive / negative cash balances
                             instruments_carry: pd.DataFrame = None,  # on nav
                             rebalancing_costs: Union[float, pd.Series] = None,  # rebalancing costs in bp
                             weight_implementation_lag: Optional[int] = None,  # applies for weight is pd.Dataframe
                             constant_trade_level: float = None,
                             is_rebalanced_at_first_date: bool = False,
                             ticker: str = 'Portfolio'  # default ticker
                             ) -> PortfolioData:
    """
    simulate portfolio given prices and weights
    include_start_date if index rebalanced at start date
    the safest weight is to pass weights as Dict or pd.Dataframe - this enforces the alignment with prices
    does not rebalance_freq when dates are pd.DataFrmae
    """
    if not isinstance(prices, pd.DataFrame):
        raise ValueError(f"prices type={type(prices)} must be pd.Dataframe")
    if isinstance(weights, pd.Series):  # map to dict
        weights = weights.to_dict()

    if isinstance(weights, Dict):  # map to np
        qu.assert_list_subset(large_list=prices.columns.to_list(),
                              list_sample=list(weights.keys()),
                              message=f"weights columns must be aligned with price columns")
        weights = prices.columns.map(weights).to_numpy()
    elif isinstance(weights, List):
        weights = np.array(weights)

    # align weights with prices
    if isinstance(weights, np.ndarray):
        if weights.shape[0] != len(prices.columns):
            raise ValueError(f"number of weights must be aligned with number of price columns")
        if len(weights.shape) > 1:
            raise ValueError(f"only single aray is allowed")

        is_rebalancing = qu.generate_rebalancing_indicators(df=prices,
                                                            freq=rebalance_freq,
                                                            include_start_date=is_rebalanced_at_first_date)

        portfolio_rebalance_dates = is_rebalancing[is_rebalancing == True]
        portfolio_weights = pd.DataFrame(data=qu.repeat_by_rows(weights, n=len(portfolio_rebalance_dates)),
                                         index=portfolio_rebalance_dates,
                                         columns=prices.columns)

    elif isinstance(weights, pd.DataFrame):
        qu.assert_list_subset(large_list=prices.columns.to_list(),
                              list_sample=weights.columns.to_list(),
                              message=f"weights columns must be aligned with price columns")
        if prices.index[0] > weights.index[0]:
            raise ValueError(f"price dates {prices.index[0]} are after weights start date {weights.index[0]}")
        portfolio_weights = weights[prices.columns]  # alighn

        # implementation lag is only valid for quant-generated weights
        if weight_implementation_lag is not None and weight_implementation_lag > 0:
            rebalancing_index = portfolio_weights.index + pd.Timedelta(days=weight_implementation_lag)
        else:
            rebalancing_index = portfolio_weights.index
        is_rebalancing = qu.set_rebalancing_timeindex_on_given_timeindex(given_index=prices.index,
                                                                         rebalancing_index=rebalancing_index)
    else:
        raise NotImplementedError(f"unsupported weights type = {type(weights)}")

    # adjust rates at rebealncing
    if funding_rate is not None:
        funding_rate_dt = qu.multiply_df_by_dt(df=funding_rate, dates=prices.index, lag=0)
    else:
        funding_rate_dt = pd.Series(0.0, index=prices.index)
    if instruments_carry is not None:
        instruments_carry_dt = qu.multiply_df_by_dt(df=instruments_carry, dates=prices.index, lag=0)
    else:
        instruments_carry_dt = pd.Series(0.0, index=prices.index)

    if rebalancing_costs is not None:
        if isinstance(rebalancing_costs, pd.Series):
            rebalancing_costs = rebalancing_costs[prices.columns].to_numpy()

    nav, units, effective_weights, realized_costs = backtest_rebalanced_portfolio(prices=prices.to_numpy(),
                                                                                  weights=portfolio_weights.to_numpy(),
                                                                                  is_rebalancing=is_rebalancing.to_numpy(),
                                                                                  funding_rate_dt=funding_rate_dt.to_numpy(),
                                                                                  instruments_carry_dt=instruments_carry_dt.to_numpy(),
                                                                                  initial_nav=initial_nav,
                                                                                  constant_trade_level=constant_trade_level,
                                                                                  rebalancing_costs=rebalancing_costs)

    portfolio_nav = pd.Series(nav, index=prices.index)
    if ticker is not None:
        portfolio_nav = portfolio_nav.rename(ticker)

    output_portfolio_data = PortfolioData(nav=portfolio_nav,
                                          units=pd.DataFrame(units, index=prices.index, columns=prices.columns),
                                          weights=pd.DataFrame(effective_weights, index=prices.index, columns=prices.columns),
                                          input_weights=weights,
                                          is_rebalancing=is_rebalancing,
                                          prices=prices,
                                          realized_costs=pd.DataFrame(realized_costs, index=prices.index, columns=prices.columns),
                                          ticker=ticker)

    return output_portfolio_data


@njit
def backtest_rebalanced_portfolio(prices: np.ndarray,
                                  weights: np.ndarray,
                                  is_rebalancing: np.ndarray,
                                  funding_rate_dt: np.ndarray = None,
                                  instruments_carry_dt: np.ndarray = None,
                                  initial_nav: float = 100.0,
                                  constant_trade_level: float = None,
                                  rebalancing_costs: Union[float, np.ndarray] = None  # proportional rebalancing costs
                                  ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    if prices.shape[0] != is_rebalancing.shape[0]:
        raise ValueError(f"prices.shape[0] != is_rebalancing.shape[0]")
    if weights.shape[1] != prices.shape[1]:
        raise ValueError(f"weights.shape[1] != prices.shape[1]")

    if funding_rate_dt is None:
        funding_rate_dt = np.zeros(prices.shape[0])

    # initialize
    current_rebalancing_idx = 0
    units = np.zeros_like(prices)
    realized_costs = np.zeros_like(prices)
    nav = np.zeros(prices.shape[0])
    cash_balances = np.zeros(prices.shape[0])

    # build initial portfolio
    current_prices = prices[0, :]
    if is_rebalancing[0]:
        if constant_trade_level is not None:
            current_units = (constant_trade_level * weights[current_rebalancing_idx]) / current_prices
        else:
            current_units = (initial_nav * weights[current_rebalancing_idx]) / current_prices

        current_units[np.isnan(current_units)] = 0
        current_cash_balance = initial_nav - np.nansum(current_units * current_prices)
        current_rebalancing_idx += 1
    else:
        current_units = np.zeros(prices.shape[1])
        current_cash_balance = initial_nav
    units[0, :] = current_units
    nav[0] = np.nansum(current_units * current_prices) + current_cash_balance  # need to be adjusted when cost are present for is_rebalancing[0] = True
    cash_balances[0] = current_cash_balance

    # loop over t
    for t in np.arange(1, prices.shape[0]):
        current_units = units[t - 1]
        current_prices = prices[t, :]
        current_cash_balance = cash_balances[t-1] * (1.0 + funding_rate_dt[t])
        
        if instruments_carry_dt is not None:
            carry = np.nansum(current_units * current_prices * instruments_carry_dt[t])
            current_cash_balance += carry

        if is_rebalancing[t]:
            if constant_trade_level:
                current_nav0 = constant_trade_level
            else:
                current_nav0 = np.nansum(current_units * current_prices) + current_cash_balance
            current_units = (current_nav0 * weights[current_rebalancing_idx]) / current_prices
            current_units[np.isnan(current_units)] = 0
            units_change = current_units - units[t-1]
            change_in_cash = -np.nansum(units_change*current_prices)
            if rebalancing_costs is not None:
                realized_costs_t = rebalancing_costs*current_prices*np.abs(units_change)
                realized_costs[t, :] = realized_costs_t
                change_in_cash -= np.nansum(realized_costs_t)
            current_cash_balance = current_cash_balance + change_in_cash
            current_rebalancing_idx += 1

        # store
        units[t, :] = current_units
        nav[t] = np.nansum(current_units * current_prices) + current_cash_balance
        cash_balances[t] = current_cash_balance

    effective_weights = np.divide(units * prices, qu.repeat_by_columns(a=nav, n=prices.shape[1]))

    return nav, units, effective_weights, realized_costs


class UnitTests(Enum):
    BLENDED = 1
    COSTS = 2


def run_unit_test(unit_test: UnitTests):

    import matplotlib.pyplot as plt
    import qis.plots.derived.prices as ppd

    from qis.test_data import load_etf_data
    prices = load_etf_data().dropna()

    prices = prices[['SPY', 'TLT']]
    # prices.iloc[:200, :] = np.nan
    print(prices)
    
    if unit_test == UnitTests.BLENDED:

        portfolio_nav_1_0 = backtest_model_portfolio(prices=prices,
                                                     weights=np.array([1.0, 0.0]),
                                                     rebalance_freq='QE').get_portfolio_nav()

        portfolio_nav_5_5 = backtest_model_portfolio(prices=prices,
                                                     weights=np.array([1.0, 0.5]),
                                                     rebalance_freq='QE').get_portfolio_nav()

        portfolio_nav_0_1 = backtest_model_portfolio(prices=prices,
                                                     weights=np.array([1.0, 1.0]),
                                                     rebalance_freq='QE').get_portfolio_nav()

        portfolio_nav = pd.concat([portfolio_nav_1_0, portfolio_nav_5_5, portfolio_nav_0_1], axis=1)
        portfolio_nav.columns = ['x1=100, x2=0', 'x1=100, x2=50', 'x1=100, x2=100']
        print(portfolio_nav)
        fig, ax = plt.subplots(1, 1, figsize=(12, 10))
        ppd.plot_prices(prices=portfolio_nav, ax=ax)

    elif unit_test == UnitTests.COSTS:
        portfolio_nav = backtest_model_portfolio(prices=prices,
                                                 weights=np.array([1.0, 1.0]),
                                                 rebalance_freq='QE')

        portfolio_nav.plot_pnl()

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.BLENDED

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



# biult in
import pandas as pd
import matplotlib.pyplot as plt
from typing import Optional, Union, List, Tuple
from enum import Enum

# qis
import qis.utils.dates as da
import qis.utils.df_ops as dfo
import qis.utils.struct_ops as sop
import qis.perfstats.perf_stats as pt
from qis.models.stats.rolling_stats import RollingPerfStat, compute_rolling_perf_stat
from qis.perfstats.config import PerfStat, PerfParams
from qis.perfstats.regime_classifier import BenchmarkReturnsQuantileRegimeSpecs

import qis.plots.derived.drawdowns as dra
import qis.plots.time_series as pts
import qis.plots.utils as put
from qis.plots.derived.regime_data import add_bnb_regime_shadows


class PerfStatsLabels(Enum):
    """
    enumerate some combinations for perf stat labels
    can refer through value
    perf_stats_labels=qis.PerfStatsLabels.DETAILED_WITH_DD.value
    perf_stats_labels=qis.PerfStatsLabels.DETAILED_WITH_DDVOL.value
    perf_stats_labels=qis.PerfStatsLabels.TOTAL_DETAILED.value
    """
    SHARPE = (PerfStat.SHARPE_RF0, )
    DETAILED_EXCESS_SHARPE = (PerfStat.PA_RETURN, PerfStat.VOL, PerfStat.SHARPE_EXCESS, )
    DETAILED_SHARPE_RF0 = (PerfStat.PA_RETURN, PerfStat.VOL, PerfStat.SHARPE_RF0,)
    DETAILED_WITH_DD = (PerfStat.PA_RETURN, PerfStat.VOL, PerfStat.SHARPE_RF0, PerfStat.MAX_DD, )
    DETAILED_WITH_SKEW = (PerfStat.PA_RETURN, PerfStat.VOL, PerfStat.SHARPE_RF0, PerfStat.SKEWNESS, )
    DETAILED_LOG_SHARPE = (PerfStat.AN_LOG_RETURN, PerfStat.VOL, PerfStat.SHARPE_LOG_AN, )
    DETAILED_WITH_DDVOL = (PerfStat.PA_RETURN, PerfStat.VOL, PerfStat.SHARPE_EXCESS, PerfStat.MAX_DD, PerfStat.MAX_DD_VOL,)
    TOTAL = (PerfStat.TOTAL_RETURN, PerfStat.MAX_DD, )
    TOTAL_DETAILED = (PerfStat.TOTAL_RETURN, PerfStat.PA_RETURN, PerfStat.VOL, PerfStat.SHARPE_EXCESS, PerfStat.MAX_DD, )


def get_performance_labels_for_stats(prices: Union[pd.DataFrame, pd.Series],
                                     perf_stats_labels: List[PerfStat] = (PerfStat.PA_RETURN, PerfStat.VOL,
                                                                          PerfStat.SHARPE_RF0, PerfStat.MAX_DD,),
                                     perf_params: PerfParams = None,
                                     **kwargs
                                     ) -> List[str]:

    if any(prices.columns.duplicated()):
        raise ValueError(f"dublicated columns:\n{prices.columns[prices.columns.duplicated()]}")

    ra_perf_table = pt.compute_ra_perf_table(prices=prices, perf_params=perf_params)
    legend_labels = []
    for index in ra_perf_table.index:
        name = index if isinstance(index, str) else str(index)
        label = f"{name}: "
        for perf_stat in perf_stats_labels:
            # pefromance strat is always defines using perf_stat.to_str()
            label += f"{perf_stat.to_str(**kwargs)}={perf_stat.to_format(**kwargs).format(ra_perf_table.loc[index, perf_stat.to_str()])}, "
        legend_labels.append(label[:-2])  # remove last ", "

    return legend_labels


def plot_prices(prices: Union[pd.DataFrame, pd.Series],
                perf_stats_labels: List[PerfStat] = (PerfStat.PA_RETURN, PerfStat.VOL, PerfStat.SHARPE_RF0, ),
                perf_params: PerfParams = None,
                regime_benchmark_str: str = None,  # to add regimes
                pivot_prices: pd.Series = None,
                regime_params: BenchmarkReturnsQuantileRegimeSpecs = BenchmarkReturnsQuantileRegimeSpecs(),
                var_format: str = '{:,.1f}',
                digits_to_show: int = 1,
                sharpe_format: str = '{:.2f}',
                x_date_freq: str = 'QE',
                trend_line: put.TrendLine = put.TrendLine.NONE,
                is_log: bool = False,
                resample_freq: str = None,
                start_to_one: bool = True,
                end_to_one: bool = False,
                title: str = None,
                ax: plt.Subplot = None,
                **kwargs
                ) -> plt.Figure:

    if isinstance(prices, pd.Series):
        prices = prices.to_frame()

    legend_labels = get_performance_labels_for_stats(prices=prices,
                                                     perf_stats_labels=perf_stats_labels,
                                                     perf_params=perf_params,
                                                     digits_to_show=digits_to_show,
                                                     sharpe_format=sharpe_format,
                                                     **kwargs)

    if resample_freq is not None:
        prices = prices.asfreq(resample_freq, method='ffill')

    if end_to_one:
        scaler = 1.0 / dfo.get_last_nonnan_values(df=prices)
        prices = prices.multiply(scaler)
    elif start_to_one:
        prices = prices.divide(dfo.get_first_nonnan_values(df=prices))

    fig = pts.plot_time_series(df=prices,
                               trend_line=trend_line,
                               var_format=var_format,
                               title=title,
                               legend_labels=legend_labels,
                               is_log=is_log,
                               x_date_freq=x_date_freq,
                               ax=ax,
                               **kwargs)

    if regime_benchmark_str is not None and regime_params is not None:
        add_bnb_regime_shadows(ax=ax,
                               data_df=prices,
                               pivot_prices=pivot_prices,
                               benchmark=regime_benchmark_str,
                               regime_params=regime_params)
    return fig


def plot_prices_with_dd(prices: Union[pd.DataFrame, pd.Series],
                        perf_stats_labels: List[PerfStat] = (PerfStat.PA_RETURN, PerfStat.VOL,
                                                             PerfStat.SHARPE_RF0),
                        perf_params: PerfParams = None,
                        regime_benchmark_str: str = None,  # to add regimes
                        pivot_prices: pd.Series = None,
                        regime_params: BenchmarkReturnsQuantileRegimeSpecs = BenchmarkReturnsQuantileRegimeSpecs(),
                        var_format: str = '{:,.1f}',
                        dd_format: str = '{:.0%}',
                        digits_to_show: int = 1,
                        start_to_one: bool = True,
                        sharpe_format: str = '{:.2f}',
                        x_date_freq: str = 'YE',
                        is_log: bool = False,
                        remove_xticklabels_ax1: bool = True,
                        title: str = 'Performance',
                        dd_title: str = 'Running Drawdown',
                        dd_legend_type: dra.DdLegendType = dra.DdLegendType.SIMPLE,
                        axs: List[plt.Subplot] = None,
                        **kwargs
                        ) -> plt.Figure:

    if isinstance(prices, pd.Series):
        prices = prices.to_frame()

    if axs is None:
        fig, axs = plt.subplots(2, 1)
    else:
        fig = None

    plot_prices(prices=prices,
                perf_params=perf_params,
                var_format=var_format,
                digits_to_show=digits_to_show,
                sharpe_format=sharpe_format,
                perf_stats_labels=perf_stats_labels,
                x_date_freq=x_date_freq,
                title=title,
                start_to_one=start_to_one,
                is_log=is_log,
                ax=axs[0],
                **kwargs)

    dra.plot_rolling_drawdowns(prices=prices,
                               perf_params=perf_params,
                               dd_legend_type=dd_legend_type,
                               x_date_freq=x_date_freq,
                               var_format=dd_format,
                               title=dd_title,
                               ax=axs[1],
                               **kwargs)

    if remove_xticklabels_ax1:
        axs[0].set_xticklabels('')

    if (regime_benchmark_str is not None or pivot_prices is not None) and regime_params is not None:
        for ax in axs:
            add_bnb_regime_shadows(ax=ax,
                                   data_df=prices,
                                   pivot_prices=pivot_prices,
                                   benchmark=regime_benchmark_str,
                                   regime_params=regime_params,
                                   perf_params=perf_params)
    return fig


def plot_prices_with_fundamentals(prices: Union[pd.DataFrame, pd.Series],
                                  volumes: Union[pd.DataFrame, pd.Series],
                                  mcap: Union[pd.DataFrame, pd.Series],
                                  perf_stats_labels: List[PerfStat] = (PerfStat.PA_RETURN, PerfStat.VOL, PerfStat.SHARPE_RF0, ),
                                  perf_params: PerfParams = None,
                                  regime_benchmark_str: str = None,  # to add regimes
                                  pivot_prices: pd.Series = None,
                                  regime_params: BenchmarkReturnsQuantileRegimeSpecs = BenchmarkReturnsQuantileRegimeSpecs(),
                                  trend_line: put.TrendLine = put.TrendLine.AVERAGE,
                                  var_format: str = '{:,.1f}',
                                  digits_to_show: int = 2,
                                  sharpe_format: str = '{:.2f}',
                                  is_log: bool = False,
                                  title: str = None,
                                  dd_title: str = 'Running Drawdown',
                                  dd_legend_type: dra.DdLegendType = dra.DdLegendType.NONE,
                                  axs: List[plt.Subplot] = None,
                                  **kwargs
                                  ) -> plt.Figure:

    if isinstance(prices, pd.Series):
        prices = prices.to_frame()

    if axs is None:
        fig, axs = plt.subplots(3, 1)
    else:
        fig = None

    plot_prices(prices=prices,
                perf_params=perf_params,
                var_format=var_format,
                digits_to_show=digits_to_show,
                sharpe_format=sharpe_format,
                perf_stats_labels=perf_stats_labels,
                is_log=is_log,
                ax=axs[0],
                **kwargs)

    dra.plot_rolling_drawdowns(prices=prices,
                               perf_params=perf_params,
                               dd_legend_type=dd_legend_type,
                               title=dd_title,
                               ax=axs[1],
                               **kwargs)

    pts.plot_time_series_2ax(df1=volumes,
                             df2=mcap,
                             trend_line1=trend_line,
                             trend_line2=trend_line,
                             var_format=var_format,
                             title=title,
                             ax=axs[2],
                             **kwargs)

    axs[0].set_xticklabels('')
    axs[1].set_xticklabels('')

    if regime_benchmark_str is not None and regime_params is not None:
        for ax in axs:
            add_bnb_regime_shadows(ax=ax,
                                   data_df=prices,
                                   pivot_prices=pivot_prices,
                                   benchmark=regime_benchmark_str,
                                   regime_params=regime_params,
                                   perf_params=perf_params)
    return fig


def plot_prices_2ax(prices_ax1: Union[pd.DataFrame, pd.Series],
                    prices_ax2: Union[pd.DataFrame, pd.Series],
                    perf_stats_labels: List[PerfStat] = (PerfStat.PA_RETURN, PerfStat.VOL, PerfStat.SHARPE_RF0),
                    perf_params: PerfParams = None,
                    var_format: str = '{:,.1f}',
                    digits_to_show: int = 2,
                    sharpe_format: str = '{:.2f}',
                    trend_line: put.TrendLine = put.TrendLine.NONE,
                    is_logs: Tuple[bool, bool] = (False, False),
                    start_to_one: bool = True,
                    end_to_one: bool = False,
                    title: str = None,
                    ax: plt.Subplot = None,
                    **kwargs
                    ) -> plt.Figure:

    if isinstance(prices_ax1, pd.Series):
        prices_ax1 = prices_ax1.to_frame()
    if isinstance(prices_ax2, pd.Series):
        prices_ax2 = prices_ax2.to_frame()

    prices_ax1.columns = [f"{x} (left)" for x in prices_ax1.columns]
    prices_ax2.columns = [f"{x} (right)" for x in prices_ax2.columns]
    legend_labels1 = get_performance_labels_for_stats(prices=prices_ax1,
                                                      perf_stats_labels=perf_stats_labels,
                                                      perf_params=perf_params,
                                                      digits_to_show=digits_to_show,
                                                      sharpe_format=sharpe_format,
                                                      **kwargs)
    legend_labels2 = get_performance_labels_for_stats(prices=prices_ax2,
                                                      perf_stats_labels=perf_stats_labels,
                                                      perf_params=perf_params,
                                                      digits_to_show=digits_to_show,
                                                      sharpe_format=sharpe_format,
                                                      **kwargs)

    legend_labels = sop.to_flat_list(legend_labels1 + legend_labels2)

    if start_to_one:
        prices_ax1 = prices_ax1.divide(dfo.get_first_nonnan_values(df=prices_ax1))

    fig = pts.plot_time_series_2ax(df1=prices_ax1,
                                   df2=prices_ax2,
                                   trend_line=trend_line,
                                   var_format=var_format,
                                   title=title,
                                   legend_labels=legend_labels,
                                   is_logs=is_logs,
                                   ax=ax,
                                   **kwargs)
    return fig


def plot_rolling_perf_stat(prices: Union[pd.Series, pd.DataFrame],
                           rolling_perf_stat: RollingPerfStat = RollingPerfStat.SHARPE,
                           time_period: da.TimePeriod = None,
                           roll_periods: int = 260,
                           roll_freq: str = None,
                           legend_stats: pts.LegendStats = pts.LegendStats.AVG_LAST,
                           title: Optional[str] = None,
                           regime_benchmark_str: str = None,
                           pivot_prices: pd.Series = None,
                           regime_params: BenchmarkReturnsQuantileRegimeSpecs = BenchmarkReturnsQuantileRegimeSpecs(),
                           perf_params: PerfParams = None,
                           ax: plt.Subplot = None,
                           **kwargs
                           ) -> plt.Figure:
    """
    plot rolling performance
    """
    df, title_ps = compute_rolling_perf_stat(prices=prices,
                                             rolling_perf_stat=rolling_perf_stat,
                                             roll_periods=roll_periods,
                                             roll_freq=roll_freq)

    if time_period is not None:
        df = time_period.locate(df)

    fig = pts.plot_time_series(df=df,
                               legend_stats=legend_stats,
                               title=title or title_ps,
                               ax=ax,
                               **sop.update_kwargs(kwargs, dict(var_format=rolling_perf_stat.value[1])))

    if regime_benchmark_str is not None and regime_params is not None:
        add_bnb_regime_shadows(ax=ax,
                               data_df=prices.reindex(index=df.index, method='ffill'),
                               pivot_prices=pivot_prices,
                               benchmark=regime_benchmark_str,
                               regime_params=regime_params,
                               perf_params=perf_params)

    return fig


class UnitTests(Enum):
    PERFORMANCE_LABELS = 1
    PRICE = 2
    PRICE_WITH_DD = 3


def run_unit_test(unit_test: UnitTests):

    from qis.test_data import load_etf_data
    prices = load_etf_data().dropna()

    if unit_test == UnitTests.PERFORMANCE_LABELS:
        this = get_performance_labels_for_stats(prices=prices, perf_stats_labels=[PerfStat.PA_RETURN,
                                                                                  PerfStat.VOL,
                                                                                  PerfStat.SHARPE_RF0,
                                                                                  PerfStat.MAX_DD])
        print(this)

    elif unit_test == UnitTests.PRICE:
        perf_params = PerfParams(freq='B')
        plot_prices(prices=prices, perf_params=perf_params)

    elif unit_test == UnitTests.PRICE_WITH_DD:
        perf_params = PerfParams(freq='ME')
        plot_prices_with_dd(prices=prices,
                            regime_benchmark_str=prices.columns[0],
                            perf_params=perf_params)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.PERFORMANCE_LABELS

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
configuration of performance stats and regime params
"""
from __future__ import annotations

import pandas as pd
from dataclasses import dataclass
from typing import Optional
from enum import Enum

from qis.utils.generic import ColVar, ValueType


class ReturnTypes(Enum):
    RELATIVE = 'Relative'  # =S1/S0-1
    LOG = 'Log'  # =ln(S1/S0)
    DIFFERENCE = 'Diff'  # =S1-S0
    LEVEL = 'Level'  # =S1
    LEVEL0 = 'Level0'  # =S0


class RegimeType(Enum):
    RETURNS = 'returns'
    LEVELS = 'levels'
    AVG = 'averages'


class RegimeData(Enum):
    REGIME_AVG = 'Average'
    REGIME_PA = 'P.a.'
    REGIME_SHARPE = 'Sharpe'


class PerfStat(ColVar, Enum):
    """
    performance data outputs are computed as df(index=assets, columns=PerfStat)
    """
    START_DATE = ColVar(name='Start date', short_n='Start\ndate', value_type=ValueType.DATE)
    END_DATE = ColVar(name='End date', short_n='End\ndate', value_type=ValueType.DATE)
    START_PRICE = ColVar(name='Start', short_n='Start\nprice', value_type=ValueType.PRICE)
    END_PRICE = ColVar(name='End', short_n='End\nprice', value_type=ValueType.PRICE)
    NUM_OBS = ColVar(name='Num Obs', short_n='Num\nObs', value_type=ValueType.INT)

    # computed in compute_pa_return_dict
    TOTAL_RETURN = ColVar(name='Total', short_n='Total return', value_type=ValueType.PERCT)
    PA_RETURN = ColVar(name='P.a. return', short='P.a.', short_n='P.a.\nreturn', value_type=ValueType.PERCT)
    AN_LOG_RETURN = ColVar(name='An. log return', short_n='An. log-return', value_type=ValueType.PERCT)
    AN_LOG_RETURN_EXCESS = ColVar(name='An. log return ex', short_n='An. log-return ex', value_type=ValueType.PERCT)
    AVG_AN_RETURN = ColVar(name='Avg. An return', short_n='Avg. An return', value_type=ValueType.PERCT)
    APR = ColVar(name='APR', short_n='APR', value_type=ValueType.PERCT)
    PA_EXCESS_RETURN = ColVar(name='P.a. excess return', short_n='P.a.\nexcess', value_type=ValueType.PERCT)
    NAV1 = ColVar(name='1$ Invested', short_n='1$ Invested', value_type=ValueType.FLOAT)
    NUM_YEARS = ColVar(name='Num Years', short_n='Num\nYears', value_type=ValueType.FLOAT)

    VOL = ColVar(name='Vol', short_n='An. vol', value_type=ValueType.PERCT)
    DOWNSIDE_VOL = ColVar(name='DownVol', short_n='DownVol', value_type=ValueType.PERCT)
    AVG_LOG_RETURN = ColVar(name='AvgLogReturn', short_n='AvgReturn', value_type=ValueType.PERCT)
    SHARPE_RF0 = ColVar(name='Sharpe (rf=0)', short='Sharpe(rf=0)', short_n='Sharpe\n(rf=0)', value_type=ValueType.SHARPE)  # compounded returns with rate = 0.0
    SHARPE_EXCESS = ColVar(name='Ex. Sharpe', short='Ex.Sharpe', short_n='Excess\nSharpe', value_type=ValueType.SHARPE)  # compunded with given rate
    SHARPE_LOG_AN = ColVar(name='An. Log Sharpe', short_n='An. Log Sharpe', value_type=ValueType.SHARPE)  # log return
    SHARPE_AVG = ColVar(name='Sharpe Avg', short_n='Sharpe Avg', value_type=ValueType.SHARPE)  # using avg return
    SHARPE_LOG_EXCESS = ColVar(name='Log Sharpe Ex', short_n='Log Sharpe Ex', value_type=ValueType.SHARPE)
    SHARPE_APR = ColVar(name='APR Sharpe', short_n='APR\nSharpe', value_type=ValueType.SHARPE)
    MARGINAL_SHARPE = ColVar(name='Marginal Sharpe', short_n='Marginal\nSharpe', value_type=ValueType.SHARPE)
    MARGINAL_SHARPE_RATIO = ColVar(name='Marginal Sharpe Ratio', short_n='Marginal\nSharpe Ratio', value_type=ValueType.FLOAT)

    SORTINO_RATIO = ColVar(name='Sortino', short_n='Sortino', value_type=ValueType.SHARPE)
    CALMAR_RATIO = ColVar(name='Calmar', short_n='Calmar', value_type=ValueType.SHARPE)

    MAX_DD = ColVar(name='Max DD', short='MaxDD', short_n='Max DD', value_type=ValueType.PERCT0)
    MAX_DD_VOL = ColVar(name='Max DD/Vol', short='MaxDD/Vol', short_n='Max DD\n/Vol', value_type=ValueType.FLOAT)
    SKEWNESS = ColVar(name='Skewness', short='Skew', short_n='Skew', value_type=ValueType.FLOAT)
    KURTOSIS = ColVar(name='Kurtosis', short_n='Kurt', value_type=ValueType.FLOAT)
    NORMTEST = ColVar(name='P-val', short_n='P-val', value_type=ValueType.FLOAT4)
    WORST = ColVar(name='Worst', short_n='Worst', value_type=ValueType.PERCT)
    BEST = ColVar(name='Best', short_n='Best', value_type=ValueType.PERCT)
    POSITIVE = ColVar(name='Positive', short_n='Positive', value_type=ValueType.PERCT)

    BEAR_AVG = ColVar(name='Bear Avg', short_n='Bear\nAvg', value_type=ValueType.PERCT)
    NORMAL_AVG = ColVar(name='Normal Avg', short_n='Normal\nAvg', value_type=ValueType.PERCT)
    BULL_AVG = ColVar(name='Bull Avg', short_n='Bull\nAvg', value_type=ValueType.PERCT)

    BEAR_PA = ColVar(name='Bear P.a.', short_n='Bear\nP.a.', value_type=ValueType.PERCT)
    NORMAL_PA = ColVar(name='Normal P.a.', short_n='Normal\nP.a.', value_type=ValueType.PERCT)
    BULL_PA = ColVar(name='Bull P.a.', short_n='Bull\nP.a.', value_type=ValueType.PERCT)

    BEAR_SHARPE = ColVar(name='Bear Sharpe', short_n='Bear\nSharpe', value_type=ValueType.SHARPE)
    NORMAL_SHARPE = ColVar(name='Normal Sharpe', short_n='Normal\nSharpe', value_type=ValueType.SHARPE)
    BULL_SHARPE = ColVar(name='Bull Sharpe', short_n='Bull\nSharpe', value_type=ValueType.SHARPE)

    BEAR_CORR = ColVar(name='Bear Correlation', short_n='Bear\nCorrelation', value_type=ValueType.FLOAT)
    NORMAL_CORR = ColVar(name='Normal Correlation', short_n='Normal\nCorrelation', value_type=ValueType.FLOAT)
    BULL_CORR = ColVar(name='Bull Correlation', short_n='Bull\nCorrelation', value_type=ValueType.FLOAT)
    TOTAL_CORR = ColVar(name='Total Correlation', short_n='Total\nCorrelation', value_type=ValueType.FLOAT)

    AVG = ColVar(name='Avg', short_n='Avg', value_type=ValueType.FLOAT)
    T_STAT = ColVar(name='T-stat', short_n='T-stat', value_type=ValueType.FLOAT)
    STD = ColVar(name='Std', short_n='Std', value_type=ValueType.FLOAT)
    STD_AN = ColVar(name='Std An', short_n='Std AN', value_type=ValueType.FLOAT)
    MEDIAN = ColVar(name='Median', short_n='50-Q', value_type=ValueType.FLOAT)
    MIN = ColVar(name='Min', short_n='Min', value_type=ValueType.FLOAT)
    MAX = ColVar(name='Max', short_n='Max', value_type=ValueType.FLOAT)
    QUANT_M_1STD = ColVar(name='-1std', short_n='16-Q', value_type=ValueType.FLOAT)
    QUANT_P1_STD = ColVar(name='+1std', short_n='84-Q', value_type=ValueType.FLOAT)
    LAST = ColVar(name='Last', short_n='Last', value_type=ValueType.PERCT)
    RANK = ColVar(name='Rank', short_n='Rank', value_type=ValueType.PERCT)

    # tre and ir
    TE = ColVar(name='TE', short_n='TE', value_type=ValueType.PERCT)
    IR = ColVar(name='IR', short_n='IR', value_type=ValueType.SHARPE)

    # linear ml
    ALPHA = ColVar(name='Alpha', short_n='Alpha', value_type=ValueType.PERCT)
    ALPHA_AN = ColVar(name='An Alpha', short_n='Alpha', value_type=ValueType.PERCT)
    BETA = ColVar(name='Beta', short_n='Beta', value_type=ValueType.FLOAT2)
    R2 = ColVar(name='R2', short_n='R2', value_type=ValueType.PERCT0)


"""
Most common table columns
"""

FULL_TABLE_COLUMNS = (PerfStat.START_DATE,
                      PerfStat.END_DATE,
                      PerfStat.NUM_OBS,
                      PerfStat.AVG_LOG_RETURN,
                      PerfStat.PA_RETURN,
                      PerfStat.VOL,
                      PerfStat.SHARPE_RF0,
                      PerfStat.MAX_DD,
                      PerfStat.MAX_DD_VOL,
                      PerfStat.SKEWNESS,
                      PerfStat.KURTOSIS,
                      PerfStat.WORST,
                      PerfStat.BEST)


RA_TABLE_COLUMNS = (PerfStat.START_DATE,
                    PerfStat.END_DATE,
                    PerfStat.PA_RETURN,
                    PerfStat.VOL,
                    PerfStat.SHARPE_RF0,
                    PerfStat.MAX_DD,
                    PerfStat.MAX_DD_VOL,
                    PerfStat.SKEWNESS,
                    PerfStat.KURTOSIS,
                    PerfStat.WORST,
                    PerfStat.BEST)


SD_PERF_COLUMNS = (PerfStat.START_DATE,
                   PerfStat.END_DATE,
                   PerfStat.PA_RETURN,
                   PerfStat.VOL,
                   PerfStat.SHARPE_RF0,
                   PerfStat.BEAR_SHARPE,
                   PerfStat.NORMAL_SHARPE,
                   PerfStat.BULL_SHARPE,
                   PerfStat.MAX_DD,
                   PerfStat.MAX_DD_VOL,
                   PerfStat.SKEWNESS)


RA_TABLE_COMPACT_COLUMNS = (PerfStat.PA_RETURN,
                            PerfStat.VOL,
                            PerfStat.SHARPE_RF0,
                            PerfStat.MAX_DD,
                            PerfStat.MAX_DD_VOL,
                            PerfStat.SKEWNESS,
                            PerfStat.KURTOSIS)


TRE_TABLE_COLUMNS = (PerfStat.TOTAL_RETURN,
                     PerfStat.AVG_AN_RETURN,
                     PerfStat.TE,
                     PerfStat.IR,
                     PerfStat.MAX_DD,
                     PerfStat.MAX_DD_VOL,
                     PerfStat.SKEWNESS,
                     PerfStat.KURTOSIS,
                     PerfStat.WORST,
                     PerfStat.BEST)


@dataclass
class PerfParams:
    """
    contain key parameters for computing risk adjusted performance
    """
    freq: str = None
    freq_vol: str = 'ME'  # volatility of Sharpe
    freq_drawdown: str = 'D'
    freq_reg: str = 'QE'  # for quadratic/linear regressions
    freq_excess_return: str = 'ME'
    return_type: ReturnTypes = ReturnTypes.LOG  # for vol computation
    rates_data: Optional[pd.Series] = None  # to compute EXCESS returns
    alpha_an_factor: float = 4.0  # to annualise alpha in linear regression, linked to frequency of freq_reg
    # alpha_an_factor = 12, 4 for freq_reg='ME', 'QE' and so

    def __post_init__(self):
        if self.freq is not None:  # global parameter
            self.freq = self.freq
            self.freq_vol = self.freq
            self.freq_drawdown = self.freq_drawdown or self.freq
            self.freq_excess_return = self.freq
        else:
            self.freq = 'ME'
            self.freq_vol = self.freq_vol
            self.freq_drawdown = self.freq_drawdown or self.freq_vol
            self.freq_excess_return = self.freq_excess_return

    def print(self):
        print(f"freq_reg: {self.freq_reg}")
        print(f"freq_vol: {self.freq_vol}")
        print(f"return_type: {self.return_type.name}")
        print(f"freq_drawdown: {self.freq_drawdown}")
        print(f"freq_excess_return: {self.freq_excess_return}")
        if self.rates_data is not None:
            print(f"rate_data:\n{self.rates_data}")

    def copy(self,
             freq_reg: str = None,
             freq_vol: str = None,
             freq_drawdown: str = None,
             freq_excess_return: str = None,
             return_type: ReturnTypes = None,
             rates_data: pd.Series = None,
             **kwargs
             ) -> PerfParams:
        this_copy = PerfParams(freq_reg=freq_reg or self.freq_reg,
                               freq_vol=freq_vol or self.freq_vol,
                               freq_drawdown=freq_drawdown or self.freq_drawdown,
                               freq_excess_return=freq_excess_return or self.freq_excess_return,
                               return_type=return_type or self.return_type,
                               rates_data=rates_data if rates_data is not None else self.rates_data)
        return this_copy


class UnitTests(Enum):
    PERFORMANCE_STAT = 1


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.PERFORMANCE_STAT:
        print(PerfStat.TOTAL_RETURN)


if __name__ == '__main__':

    unit_test = UnitTests.PERFORMANCE_STAT

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)




import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import qis as qis
from bbg_fetch import fetch_field_timeseries_per_tickers


tickers = {'CICMCE5B Index': 'ITRAXX Main 5Y',
           'CICMCE1B Index': 'ITRAXX Main 10Y',
           'CICMCX5B Index': 'ITRAXX Xover 5Y',
           'CICMCI5B Index': 'CDX NA IG 5Y',
           'CICMCI1B Index': 'CDX NA IG 10Y',
           'CICMCH5B Index': 'CDX NA HY 5Y',
           'CICMCG05 Index': 'UK CDS 5Y',
           'CICMCK05 Index': 'South Korea CDS 5Y',
           'CICMCJ05 Index': 'Japan CDS 5Y',
           'CICMCU05 Index': 'US CDS 5Y',
           'CICMCE05 Index': 'Germany CDS 5Y'
           }

tickers = {'UISYMH5S Index': 'UBS_CDX_HY',
           'CICMCH5B Index': 'CITI_CDX_HY',
           'CH5LMER5 Index': 'GS_CDX_HY',
           'UISYMI5S Index': 'UBS_IG_5Y',
           'CICMCI5B Index': 'CITI_IG_5Y',
           'CI5LMER5 Index': 'GS_IG_5Y'}

prices = fetch_field_timeseries_per_tickers(tickers=tickers, freq='B', field='PX_LAST').ffill()
print(prices)
benchmark_prices = fetch_field_timeseries_per_tickers(tickers={'LGDRTRUU Index': 'Agg IG'}, freq='B', field='PX_LAST').ffill()

time_period = qis.TimePeriod('01Jan2008', pd.Timestamp.today())
kwargs = qis.fetch_default_report_kwargs(time_period=time_period, add_rates_data=False)

fig = qis.generate_multi_asset_factsheet(prices=prices,
                                         benchmark_prices=benchmark_prices,
                                         time_period=time_period,
                                         **kwargs)
qis.save_figs_to_pdf(figs=[fig],
                     file_name=f"credit_trackers_report", orientation='landscape',
                     local_path=qis.local_path.get_output_path())



"""
plot returns heatmap table by monthly and annual
"""
# packages
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from enum import Enum
from matplotlib.ticker import FuncFormatter

# qis
import qis.plots.utils as put
from qis.perfstats.regime_classifier import BenchmarkReturnsQuantileRegimeSpecs, BenchmarkReturnsQuantilesRegime


def plot_regime_pdf(prices: pd.DataFrame,
                    benchmark: str,
                    regime_params: BenchmarkReturnsQuantileRegimeSpecs = BenchmarkReturnsQuantileRegimeSpecs(),
                    ax: plt.Subplot = None,
                    var_format: str = '{:.0%}',
                    is_histogram: bool = False,
                    is_multiple_stack: bool = False,
                    title: str = None,
                    fontsize: int = 10,
                    bins: int = 30,
                    legend_loc: str = None,
                    **kwargs
                    ) -> plt.Figure:

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = None

    regime_classifier = BenchmarkReturnsQuantilesRegime(regime_params=regime_params)
    sampled_returns_with_regime_id = regime_classifier.compute_sampled_returns_with_regime_id(prices=prices,
                                                                                              benchmark=benchmark,
                                                                                              **regime_params._asdict())

    if is_histogram:
        sns.histplot(data=sampled_returns_with_regime_id,
                     x=benchmark,
                     hue=regime_classifier.REGIME_COLUMN,
                     hue_order=regime_classifier.get_regime_ids_colors().keys(),
                     multiple='stack' if is_multiple_stack else 'layer',
                     bins=bins,
                     palette=regime_classifier.get_regime_ids_colors().values(),
                     ax=ax)


    else:
        sns.kdeplot(data=sampled_returns_with_regime_id,
                    x=benchmark,
                    hue=regime_classifier.REGIME_COLUMN,
                    hue_order=regime_classifier.get_regime_ids_colors().keys(),
                    palette=regime_classifier.get_regime_ids_colors().values(),
                    ax=ax)

    ax.xaxis.set_major_formatter(FuncFormatter(lambda x, _: var_format.format(x)))

    put.set_legend(ax=ax, legend_loc=legend_loc, fontsize=fontsize, **kwargs)

    ax.get_yaxis().set_visible(False)
    put.set_spines(ax=ax, **kwargs)
    if title is not None:
        ax.set_title(label=title, **kwargs)

    return fig


class UnitTests(Enum):
    REGIME_PDF = 1


def run_unit_test(unit_test: UnitTests):

    from qis.test_data import load_etf_data
    prices = load_etf_data()[['SPY', 'TLT']].dropna()

    if unit_test == UnitTests.REGIME_PDF:
        with sns.axes_style("darkgrid"):
            fig, axs = plt.subplots(1, 2, figsize=(15, 8), tight_layout=True)
            plot_regime_pdf(prices=prices, benchmark='SPY', is_histogram=False, ax=axs[0])
            plot_regime_pdf(prices=prices, benchmark='SPY', is_histogram=True, ax=axs[1])

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.REGIME_PDF

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
generate portfolio factsheet report using PortfolioData object
PortfolioData object can be generated by a backtester or actucal strategy data
with comparison to 1-2 cash benchmarks
PortfolioData can contain either simulated or actual portfolio data
"""
import numpy as np
# packages
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from typing import Tuple, Optional, List, Union
import qis as qis
from qis import TimePeriod, PerfParams, BenchmarkReturnsQuantileRegimeSpecs
from qis.portfolio.portfolio_data import PortfolioData
from qis.portfolio.reports.config import PERF_PARAMS, REGIME_PARAMS


def generate_strategy_factsheet(portfolio_data: PortfolioData,
                                benchmark_prices: Union[pd.DataFrame, pd.Series],
                                time_period: TimePeriod = None,
                                ytd_attribution_time_period: TimePeriod = qis.get_ytd_time_period(),
                                perf_params: PerfParams = PERF_PARAMS,
                                regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                                regime_benchmark: str = None,  # default is set to benchmark_prices.columns[0]
                                exposures_freq: Optional[str] = 'W-WED',  #'W-WED',
                                turnover_rolling_period: int = 260,
                                freq_turnover: str = 'B',
                                freq_cost: str = 'B',
                                cost_rolling_period: int = 260,
                                factor_beta_span: int = 52,
                                freq_beta: str = 'W-WED',
                                vol_rolling_window: int = 13,
                                freq_sharpe: str = 'B',
                                freq_var: str = 'B',
                                var_span: float = 33,
                                sharpe_rolling_window: int = 260,
                                add_benchmarks_to_navs: bool = False,
                                figsize: Tuple[float, float] = (8.5, 11.7),  # A4 for portrait
                                fontsize: int = 4,
                                weight_change_sample_size: int = 20,
                                weight_report_time_period: TimePeriod = None,
                                add_current_position_var_risk_sheet: bool = True,
                                add_weights_turnover_sheet: bool = True,
                                add_grouped_exposures: bool = False,
                                add_grouped_cum_pnl: bool = False,
                                add_weight_change_report: bool = False,
                                add_current_signal_report: bool = False,
                                add_instrument_history_report: bool = False,
                                y_limits_signal: Tuple[Optional[float], Optional[float]] = (-1.0, 1.0),
                                is_1y_exposures: bool = False,
                                is_grouped: Optional[bool] = None,
                                dd_legend_type: qis.DdLegendType = qis.DdLegendType.SIMPLE,
                                is_norm_costs: bool = True,
                                df_to_add: pd.DataFrame = None,
                                factsheet_name: str = None,
                                **kwargs
                                ) -> List[plt.Figure]:
    # align
    if isinstance(benchmark_prices, pd.Series):
        benchmark_prices = benchmark_prices.to_frame()

    benchmark_prices = benchmark_prices.reindex(index=portfolio_data.nav.index, method='ffill')
    if regime_benchmark is None:
        regime_benchmark = benchmark_prices.columns[0]

    if is_grouped is None:
        if len(portfolio_data.get_weights().columns) >= 10:  # otherwise tables look too bad
            is_grouped = True
        else:
            is_grouped = False

    fig = plt.figure(figsize=figsize, constrained_layout=True)
    gs = fig.add_gridspec(nrows=14, ncols=4, wspace=0.0, hspace=0.0)

    plot_kwargs = dict(fontsize=fontsize,
                       linewidth=0.5,
                       digits_to_show=1, sharpe_digits=2,
                       weight='normal',
                       markersize=1,
                       framealpha=0.75)
    kwargs = qis.update_kwargs(kwargs, plot_kwargs)
    if factsheet_name is None:
        factsheet_name = f"{portfolio_data.nav.name} factsheet"
    fig.suptitle(factsheet_name, fontweight="bold", fontsize=8, color='blue')

    # prices
    portfolio_nav = portfolio_data.get_portfolio_nav(time_period=time_period)
    # set reporting time period here
    if time_period is None:
        time_period = qis.get_time_period(portfolio_nav)

    if add_benchmarks_to_navs:
        benchmark_prices_ = benchmark_prices
        joint_prices = pd.concat([portfolio_nav, benchmark_prices_], axis=1).dropna()
        pivot_prices = joint_prices[regime_benchmark]
    else:
        benchmark_prices_ = benchmark_prices[regime_benchmark]
        joint_prices = pd.concat([portfolio_nav, benchmark_prices_], axis=1).dropna()
        pivot_prices = joint_prices[regime_benchmark]
        joint_prices = joint_prices[portfolio_nav.name]

    ax = fig.add_subplot(gs[0:2, :2])
    qis.plot_prices(prices=joint_prices,
                    perf_params=perf_params,
                    title=f"Cumulative performance with background colors using bear/normal/bull regimes of {regime_benchmark} {regime_params.freq}-returns",
                    ax=ax,
                    **kwargs)
    qis.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)
    qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)

    # dd
    ax = fig.add_subplot(gs[2:4, :2])
    qis.plot_rolling_drawdowns(prices=joint_prices,
                               title='Running Drawdowns',
                               dd_legend_type=dd_legend_type,
                               ax=ax, **kwargs)
    qis.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)
    qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)

    # under watre
    ax = fig.add_subplot(gs[4:6, :2])
    qis.plot_rolling_time_under_water(prices=joint_prices,
                                      title='Running Time under Water',
                                      dd_legend_type=dd_legend_type,
                                      ax=ax, **kwargs)
    qis.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)
    qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)

    # rolling performance
    ax = fig.add_subplot(gs[6:8, :2])
    portfolio_data.plot_rolling_perf(rolling_perf_stat=qis.RollingPerfStat.TOTAL_RETURNS,
                                     add_benchmarks=add_benchmarks_to_navs,
                                     roll_freq=freq_sharpe,
                                     rolling_window=sharpe_rolling_window,
                                     time_period=time_period,
                                     ax=ax,
                                     **qis.update_kwargs(kwargs, dict(trend_line=qis.TrendLine.ZERO_SHADOWS)))
    qis.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)
    qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)

    # exposures
    exposures = portfolio_data.get_weights(is_grouped=is_grouped, time_period=time_period, add_total=False)
    ax = fig.add_subplot(gs[8:10, :2])
    if exposures_freq is not None:
        exposures = exposures.resample(exposures_freq).last()
    qis.plot_stack(df=exposures,
                   use_bar_plot=True,
                   title='Exposures',
                   legend_stats=qis.LegendStats.AVG_NONNAN_LAST,
                   var_format='{:.1%}',
                   ax=ax,
                   **qis.update_kwargs(kwargs, dict(bbox_to_anchor=(0.5, 1.05), ncols=2)))
    qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)

    # turnover
    ax = fig.add_subplot(gs[10:12, :2])
    turnover = portfolio_data.get_turnover(time_period=time_period, roll_period=turnover_rolling_period,
                                           freq=freq_turnover,
                                           is_grouped=is_grouped)
    freq = pd.infer_freq(turnover.index)
    turnover_title = f"{turnover_rolling_period}-period rolling {freq}-freq Turnover"
    qis.plot_time_series(df=turnover,
                         var_format='{:,.2%}',
                         # y_limits=(0.0, None),
                         legend_stats=qis.LegendStats.AVG_NONNAN_LAST,
                         title=turnover_title,
                         ax=ax,
                         **kwargs)
    qis.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)
    qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)

    # costs
    ax = fig.add_subplot(gs[12:14, :2])
    costs = portfolio_data.get_costs(time_period=time_period, roll_period=cost_rolling_period,
                                     freq=freq_cost,
                                     is_grouped=is_grouped,
                                     is_norm_costs=is_norm_costs)
    freq = pd.infer_freq(costs.index)
    costs_title = f"{cost_rolling_period}-period rolling {freq}-freq Costs"
    qis.plot_time_series(df=costs,
                         var_format='{:,.2%}',
                         # y_limits=(0.0, None),
                         legend_stats=qis.LegendStats.AVG_NONNAN_LAST,
                         title=costs_title,
                         ax=ax,
                         **kwargs)
    qis.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)
    qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)

    # ra perf table
    if add_benchmarks_to_navs:
        benchmark_price1 = benchmark_prices
    else:
        benchmark_price1 = benchmark_prices[regime_benchmark]
    if is_grouped:
        ax = fig.add_subplot(gs[:2, 2:])
        portfolio_data.plot_ra_perf_table(ax=ax,
                                          benchmark_price=benchmark_price1,
                                          time_period=time_period,
                                          perf_params=perf_params,
                                          is_grouped=is_grouped,
                                          **qis.update_kwargs(kwargs, dict(fontsize=fontsize)))
    else:  # plot two tables
        ax = fig.add_subplot(gs[0, 2:])
        portfolio_data.plot_ra_perf_table(ax=ax,
                                          benchmark_price=benchmark_price1,
                                          time_period=time_period,
                                          perf_params=perf_params,
                                          is_grouped=is_grouped,
                                          **qis.update_kwargs(kwargs, dict(fontsize=fontsize)))
        ax = fig.add_subplot(gs[1, 2:])
        # change regression to weekly
        time_period1 = qis.get_time_period_shifted_by_years(time_period=time_period)
        if pd.infer_freq(benchmark_prices.index) in ['B', 'D']:
            local_kwargs = qis.update_kwargs(kwargs, dict(time_period=time_period1, alpha_an_factor=52, freq_reg='W-WED', fontsize=fontsize))
        else:
            local_kwargs = qis.update_kwargs(kwargs, dict(time_period=time_period1, fontsize=fontsize))
        portfolio_data.plot_ra_perf_table(ax=ax,
                                          benchmark_price=benchmark_price1,
                                          perf_params=perf_params,
                                          is_grouped=is_grouped,
                                          **local_kwargs)

    # heatmap
    ax = fig.add_subplot(gs[2:4, 2:])
    portfolio_data.plot_monthly_returns_heatmap(ax=ax,
                                                time_period=time_period,
                                                title='Monthly Returns',
                                                **qis.update_kwargs(kwargs, dict(fontsize=fontsize, date_format='%Y')))

    # periodic returns
    ax = fig.add_subplot(gs[4:6, 2:])
    local_kwargs = qis.update_kwargs(kwargs=kwargs,
                                     new_kwargs=dict(fontsize=fontsize, square=False, x_rotation=90, transpose=True))
    portfolio_data.plot_periodic_returns(benchmark_prices=benchmark_price1,
                                         is_grouped=is_grouped,
                                         time_period=time_period,
                                         ax=ax,
                                         **local_kwargs)

    # regime data
    portfolio_data.plot_regime_data(is_grouped=is_grouped,
                                    benchmark_price=benchmark_prices[regime_benchmark],
                                    time_period=time_period,
                                    perf_params=perf_params,
                                    regime_params=regime_params,
                                    ax=fig.add_subplot(gs[6:8, 2:]),
                                    **kwargs)

    # perf attribution
    local_kwargs = qis.update_kwargs(kwargs=kwargs, new_kwargs=dict(legend_loc=None))
    with sns.axes_style("whitegrid"):
        portfolio_data.plot_performance_attribution(time_period=time_period,
                                                    attribution_metric=qis.AttributionMetric.PNL,
                                                    ax=fig.add_subplot(gs[8:10, 2:]),
                                                    **local_kwargs)
        portfolio_data.plot_performance_attribution(time_period=time_period,
                                                    attribution_metric=qis.AttributionMetric.PNL_RISK,
                                                    ax=fig.add_subplot(gs[10:12, 2:]),
                                                    **local_kwargs)

    # constituents
    ax = fig.add_subplot(gs[12:, 2:])
    num_investable_instruments = portfolio_data.get_num_investable_instruments(time_period=time_period)
    qis.plot_time_series(df=num_investable_instruments,
                         var_format='{:,.0f}',
                         legend_stats=qis.LegendStats.FIRST_AVG_LAST,
                         title='Number of investable and invested instruments',
                         ax=ax,
                         **kwargs)
    qis.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)
    qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)

    figs = [fig]

    if add_current_position_var_risk_sheet:
        # qqq
        fig = plt.figure(figsize=figsize, constrained_layout=True)
        fig.suptitle(f"{portfolio_data.nav.name} 99%-Var and risk profile", fontweight="bold", fontsize=8, color='blue')
        figs.append(fig)
        gs = fig.add_gridspec(nrows=7, ncols=4, wspace=0.0, hspace=0.0)

        # current var grouped
        with sns.axes_style("whitegrid"):
            axs = [fig.add_subplot(gs[0, 0]), fig.add_subplot(gs[0, 1]), fig.add_subplot(gs[0, 2]), fig.add_subplot(gs[0, 3])]
            portfolio_data.plot_current_var(snapshot_period=qis.SnapshotPeriod.LAST,
                                            is_grouped=True, is_correlated=False, time_period=time_period,
                                            freq=freq_var, vol_span=var_span,
                                            ax=axs[0], **kwargs)
            portfolio_data.plot_current_var(snapshot_period=qis.SnapshotPeriod.MAX,
                                            is_grouped=True, is_correlated=False, time_period=time_period,
                                            freq=freq_var, vol_span=var_span,
                                            ax=axs[1], **kwargs)

            portfolio_data.plot_current_var(snapshot_period=qis.SnapshotPeriod.LAST,
                                            is_grouped=True, is_correlated=True, time_period=time_period,
                                            freq=freq_var, vol_span=var_span,
                                            ax=axs[2], **kwargs)
            portfolio_data.plot_current_var(snapshot_period=qis.SnapshotPeriod.MAX,
                                            is_grouped=True, is_correlated=True, time_period=time_period,
                                            freq=freq_var, vol_span=var_span,
                                            ax=axs[3], **kwargs)
            qis.align_y_limits_axs(axs=axs)

        # last / max var by instrument
        with sns.axes_style("whitegrid"):
            axs = [fig.add_subplot(gs[1, 0:2]), fig.add_subplot(gs[1, 2:4])]
            portfolio_data.plot_current_var(snapshot_period=qis.SnapshotPeriod.LAST,
                                            is_grouped=False, is_correlated=False, time_period=time_period,
                                            freq=freq_var, vol_span=var_span,
                                            ax=axs[0], **kwargs)
            portfolio_data.plot_current_var(snapshot_period=qis.SnapshotPeriod.MAX,
                                            is_grouped=False, is_correlated=False, time_period=time_period,
                                            freq=freq_var, vol_span=var_span,
                                            ax=axs[1], **kwargs)
            qis.align_y_limits_axs(axs=axs)

        # best worst returns
        with sns.axes_style("whitegrid"):
            portfolio_data.plot_best_worst_returns(ax=fig.add_subplot(gs[2, 0:2]),
                                                   num_returns=20,
                                                   time_period=time_period,
                                                   title=f"Portfolio Worst/Best 20 returns for {time_period.to_str()}",
                                                   **qis.update_kwargs(kwargs, dict(date_format='%d%b%Y')))
            num_assets = 10
            num_investable_assets = len(num_investable_instruments.columns)
            if num_investable_assets > 2 * num_assets:
                pre_title = f"-{num_assets}"
            else:
                pre_title = ''
            portfolio_data.plot_contributors(ax=fig.add_subplot(gs[2, 2]),
                                             time_period=time_period,
                                             title=f"Bottom/Top{pre_title} performance contributors {time_period.to_str()}",
                                             **kwargs)
            time_period_1y = qis.get_time_period_shifted_by_years(time_period=time_period)
            portfolio_data.plot_contributors(ax=fig.add_subplot(gs[2, 3]),
                                             time_period=time_period_1y,
                                             title=f"Bottom/Top-{num_assets} performance contributors {time_period_1y.to_str()}",
                                             **kwargs)

        # var attribution
        with sns.axes_style("whitegrid"):
            portfolio_data.plot_var_stack(is_grouped=True, is_correlated=False,
                                          time_period=time_period,
                                          ax=fig.add_subplot(gs[3, :2]),
                                          **kwargs)
            portfolio_data.plot_var_stack(is_grouped=True, is_correlated=True,
                                          time_period=time_period,
                                          ax=fig.add_subplot(gs[3, 2:]),
                                          **kwargs)

        # var time series - independent
        ax = fig.add_subplot(gs[4, :2])
        portfolio_data.plot_portfolio_grouped_var(ax=ax,
                                                  is_correlated=False,
                                                  time_period=time_period,
                                                  **kwargs)
        qis.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)
        qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)

        # var time series - correlted
        ax = fig.add_subplot(gs[4, 2:])
        portfolio_data.plot_portfolio_grouped_var(ax=ax,
                                                  is_correlated=True,
                                                  time_period=time_period,
                                                  **kwargs)
        qis.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)
        qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)

        # vol time series
        ax = fig.add_subplot(gs[5, 2:])
        portfolio_data.plot_portfolio_vols(freq=perf_params.freq_vol,
                                           span=vol_rolling_window,
                                           time_period=time_period,
                                           ax=ax,
                                           **kwargs)
        qis.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)
        qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)

        # benchmark betas
        ax = fig.add_subplot(gs[5, :2])
        factor_exposures = portfolio_data.compute_portfolio_benchmark_betas(benchmark_prices=benchmark_prices,
                                                                            time_period=time_period,
                                                                            freq_beta=freq_beta,
                                                                            factor_beta_span=factor_beta_span)
        factor_beta_title = f"Rolling {factor_beta_span}-span beta of {freq_beta}-freq returns"
        qis.plot_time_series(df=factor_exposures,
                             var_format='{:,.2f}',
                             legend_stats=qis.LegendStats.AVG_NONNAN_LAST,
                             title=factor_beta_title,
                             ax=ax,
                             **kwargs)
        qis.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)
        qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)

        # beta attribution
        ax = fig.add_subplot(gs[6, :2])
        factor_attribution = portfolio_data.compute_portfolio_benchmark_attribution(benchmark_prices=benchmark_prices,
                                                                                    freq_beta=freq_beta,
                                                                                    factor_beta_span=factor_beta_span,
                                                                                    time_period=time_period)
        factor_attribution_title = (f"Cumulative return attribution using rolling "
                                    f"{factor_beta_span}-span beta of {freq_beta}-freq returns")
        qis.plot_time_series(df=factor_attribution.cumsum(0),
                             var_format='{:,.0%}',
                             legend_stats=qis.LegendStats.LAST_NONNAN,
                             title=factor_attribution_title,
                             ax=ax,
                             **kwargs)
        qis.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)
        qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)
        """
        # returns scatter
        with sns.axes_style("whitegrid"):
            portfolio_data.plot_returns_scatter(ax=fig.add_subplot(gs[6, 2:]),
                                                benchmark_price=benchmark_prices[regime_benchmark],
                                                time_period=time_period,
                                                freq=perf_params.freq_reg,
                                                is_grouped=is_grouped,
                                                **kwargs)

        """
        # vol regime data
        ax = fig.add_subplot(gs[6, 2:])
        portfolio_data.plot_vol_regimes(ax=ax,
                                        benchmark_price=benchmark_prices[regime_benchmark],
                                        is_grouped=is_grouped,
                                        time_period=time_period,
                                        freq=regime_params.freq,
                                        regime_params=regime_params,
                                        **kwargs)
        """
        if len(benchmark_prices.columns) > 1:
            ax = fig.add_subplot(gs[12:14, 2:])
            portfolio_data.plot_vol_regimes(ax=ax,
                                            benchmark_price=benchmark_prices.iloc[:, 1],
                                            is_grouped=is_grouped,
                                            time_period=time_period,
                                            freq=regime_params.freq,
                                            regime_params=regime_params,
                                            **kwargs)
        """
    if add_weights_turnover_sheet:
        # qqq
        fig = plt.figure(figsize=figsize, constrained_layout=True)
        fig.suptitle(f"{portfolio_data.nav.name} current position profile", fontweight="bold", fontsize=8, color='blue')
        figs.append(fig)
        gs = fig.add_gridspec(nrows=7, ncols=4, wspace=0.0, hspace=0.0)

        # current position
        with sns.axes_style("whitegrid"):
            portfolio_data.plot_current_weights(is_grouped=False,
                                                ax=fig.add_subplot(gs[0, :2]), **kwargs)
            portfolio_data.plot_current_weights(is_grouped=True,
                                                ax=fig.add_subplot(gs[0, 2:]), **kwargs)

        # change in position
        with sns.axes_style("whitegrid"):
            portfolio_data.plot_last_weights_change(is_grouped=False,
                                                    ax=fig.add_subplot(gs[1, :2]), **kwargs)
            portfolio_data.plot_last_weights_change(is_grouped=True,
                                                    ax=fig.add_subplot(gs[1, 2:]), **kwargs)

        # weights for ytd performance attribution
        with sns.axes_style("whitegrid"):
            portfolio_data.plot_current_weights(is_grouped=False,
                                                time_period=ytd_attribution_time_period,
                                                ax=fig.add_subplot(gs[2, :2]),
                                                **kwargs)
            portfolio_data.plot_current_weights(is_grouped=True,
                                                time_period=ytd_attribution_time_period,
                                                ax=fig.add_subplot(gs[2, 2:]),
                                                **kwargs)

        # total and ytd performance attribution
        with sns.axes_style("whitegrid"):
            local_kwargs = qis.update_kwargs(kwargs=kwargs, new_kwargs=dict(legend_loc=None))
            portfolio_data.plot_performance_attribution(time_period=time_period,
                                                        attribution_metric=qis.AttributionMetric.PNL,
                                                        ax=fig.add_subplot(gs[3, :2]),
                                                        **local_kwargs)
            portfolio_data.plot_performance_attribution(time_period=ytd_attribution_time_period,
                                                        attribution_metric=qis.AttributionMetric.PNL,
                                                        ax=fig.add_subplot(gs[3, 2:]),
                                                        **local_kwargs)
        # turnover
        with sns.axes_style("whitegrid"):
            local_kwargs = qis.update_kwargs(kwargs=kwargs, new_kwargs=dict(legend_loc=None))
            portfolio_data.plot_performance_attribution(time_period=time_period,
                                                        attribution_metric=qis.AttributionMetric.TURNOVER,
                                                        ax=fig.add_subplot(gs[4, :2]),
                                                        **local_kwargs)
            portfolio_data.plot_performance_attribution(time_period=ytd_attribution_time_period,
                                                        attribution_metric=qis.AttributionMetric.TURNOVER,
                                                        ax=fig.add_subplot(gs[4, 2:]),
                                                        **local_kwargs)

        # vol adjusted turnover
        with sns.axes_style("whitegrid"):
            local_kwargs = qis.update_kwargs(kwargs=kwargs, new_kwargs=dict(legend_loc=None))
            portfolio_data.plot_performance_attribution(time_period=time_period,
                                                        attribution_metric=qis.AttributionMetric.VOL_ADJUSTED_TURNOVER,
                                                        ax=fig.add_subplot(gs[5, :2]),
                                                        **local_kwargs)
            portfolio_data.plot_performance_attribution(time_period=ytd_attribution_time_period,
                                                        attribution_metric=qis.AttributionMetric.VOL_ADJUSTED_TURNOVER,
                                                        ax=fig.add_subplot(gs[5, 2:]),
                                                        **local_kwargs)

        # costs
        with sns.axes_style("whitegrid"):
            local_kwargs = qis.update_kwargs(kwargs=kwargs, new_kwargs=dict(legend_loc=None, is_norm_costs=is_norm_costs))
            portfolio_data.plot_performance_attribution(time_period=time_period,
                                                        attribution_metric=qis.AttributionMetric.COSTS,
                                                        ax=fig.add_subplot(gs[6, :2]),
                                                        **local_kwargs)
            portfolio_data.plot_performance_attribution(time_period=ytd_attribution_time_period,
                                                        attribution_metric=qis.AttributionMetric.COSTS,
                                                        ax=fig.add_subplot(gs[6, 2:]),
                                                        **local_kwargs)

    if add_current_signal_report and portfolio_data.strategy_signal_data is not None:
        fig = qis.generate_current_signal_report(portfolio_data=portfolio_data,
                                                 **qis.update_kwargs(kwargs, dict(fontsize=5, figsize=figsize,
                                                                                  y_limits=y_limits_signal)))
        figs.append(fig)

    if add_weight_change_report and portfolio_data.strategy_signal_data is not None:
        fig = qis.generate_weight_change_report(portfolio_data=portfolio_data,
                                                time_period=weight_report_time_period or time_period,
                                                sample_size=weight_change_sample_size,
                                                is_grouped=True,
                                                **qis.update_kwargs(kwargs, dict(fontsize=5, figsize=figsize)))
        figs.append(fig)

    # set 1y time period for exposures
    if is_1y_exposures:
        time_period1 = qis.get_time_period_shifted_by_years(time_period=time_period)
        regime_params1 = BenchmarkReturnsQuantileRegimeSpecs(freq='ME')
    else:
        time_period1 = weight_report_time_period or time_period
        regime_params1 = regime_params

    if add_grouped_exposures:
        grouped_exposures_agg, grouped_exposures_by_inst = portfolio_data.get_grouped_long_short_exposures(time_period=time_period1)
        nrows = len(grouped_exposures_agg.keys())
        fig1 = plt.figure(figsize=figsize, constrained_layout=True)
        figs.append(fig1)
        fig1.suptitle(f"{portfolio_data.nav.name} Exposures by groups for period {time_period1.to_str()}",
                     fontweight="bold", fontsize=8, color='blue')
        gs = fig1.add_gridspec(nrows=nrows, ncols=2, wspace=0.0, hspace=0.0)
        local_kwargs = qis.update_kwargs(kwargs=kwargs, new_kwargs=dict(framealpha=0.9))
        for idx, (group, exposures_agg) in enumerate(grouped_exposures_agg.items()):
            datas = {f"{group} aggregated": grouped_exposures_agg[group],
                     f"{group} by instrument": grouped_exposures_by_inst[group]}
            for idx_, (key, df) in enumerate(datas.items()):
                ax = fig1.add_subplot(gs[idx, idx_])
                qis.plot_time_series(df=df,
                                     var_format='{:,.0%}',
                                     legend_stats=qis.LegendStats.AVG_MIN_MAX_LAST,
                                     title=f"{key}",
                                     ax=ax,
                                     **local_kwargs)
                qis.add_bnb_regime_shadows(ax=ax, pivot_prices=time_period1.locate(pivot_prices),
                                           regime_params=regime_params1)
                qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)
                ax.axhline(0, color='black', linewidth=0.5)

    if add_grouped_cum_pnl:
        grouped_pnls_agg, grouped_pnls_by_inst = portfolio_data.get_grouped_cum_pnls(time_period=time_period1)
        nrows = len(grouped_pnls_agg.keys())
        fig1 = plt.figure(figsize=figsize, constrained_layout=True)
        figs.append(fig1)
        fig1.suptitle(f"{portfolio_data.nav.name} P&L by groups for period {time_period1.to_str()}",
                     fontweight="bold", fontsize=8, color='blue')
        gs = fig1.add_gridspec(nrows=nrows, ncols=2, wspace=0.0, hspace=0.0)
        local_kwargs = qis.update_kwargs(kwargs=kwargs, new_kwargs=dict(framealpha=0.9))
        for idx, (group, pnls_agg) in enumerate(grouped_pnls_agg.items()):
            datas = {f"{group} aggregated": grouped_pnls_agg[group],
                     f"{group} by instrument": grouped_pnls_by_inst[group]}
            for idx_, (key, df) in enumerate(datas.items()):
                ax = fig1.add_subplot(gs[idx, idx_])
                qis.plot_time_series(df=df,
                                     var_format='{:,.0%}',
                                     legend_stats=qis.LegendStats.LAST_NONNAN,
                                     title=f"{key}",
                                     ax=ax,
                                     **local_kwargs)
                qis.add_bnb_regime_shadows(ax=ax,
                                           pivot_prices=time_period1.locate(pivot_prices),
                                           regime_params=regime_params1)
                qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)

    if add_instrument_history_report:
        if df_to_add is None:
            ac_data = portfolio_data.group_data.to_frame(name='AC')
            if portfolio_data.instrument_names is not None:
                df_to_add = pd.concat([portfolio_data.instrument_names.rename('Name'),
                                       ac_data], axis=1)
            else:
                df_to_add = ac_data

        perf_columns = (qis.PerfStat.START_DATE, qis.PerfStat.END_DATE, qis.PerfStat.PA_RETURN,
                        qis.PerfStat.VOL, qis.PerfStat.SHARPE_RF0,
                        qis.PerfStat.MAX_DD, qis.PerfStat.MAX_DD_VOL, qis.PerfStat.SKEWNESS)

        fig = qis.generate_price_history_report(prices=portfolio_data.prices,
                                                **qis.update_kwargs(kwargs, dict(fontsize=4, figsize=figsize,
                                                                                 perf_columns=perf_columns,
                                                                                 df_to_add=df_to_add)))
        fig.suptitle('Program Instrument Universe', fontweight="bold", fontsize=8, color='blue')
        figs.append(fig)

    return figs



"""
heatmap plots
"""

# packages
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from typing import List, Optional
from enum import Enum

# qis
import qis.plots.utils as put


def plot_heatmap(df: pd.DataFrame,
                 transpose: bool = False,
                 inverse: bool = False,
                 date_format: Optional[str] = '%Y',
                 cmap: str = 'RdYlGn',
                 var_format: Optional[str] = '{:.1%}',
                 alpha: float = 1.0,
                 fontsize: int = 10,
                 title: Optional[str] = None,
                 top_x_label: bool = True,
                 square: bool = False,
                 vline_columns: List[int] = None,
                 hline_rows: List[int] = None,
                 vmin: float = None,
                 vmax: float = None,
                 labelpad: int = 50,
                 ylabel: str = '',
                 ax: plt.Subplot = None,
                 **kwargs
                 ) -> Optional[plt.Figure]:

    if ax is None:
        fig, ax = plt.subplots()
    else:  # add table to existing axis
        fig = None

    df = df.copy()

    if date_format is not None:  # index may include 'Total'
        df.index = [date.strftime(date_format) if isinstance(date, pd.Timestamp) else date for date in df.index]

    if transpose:
        df = df.T
        inverse = False

    if inverse:
        df = df.reindex(index=df.index[::-1])

    if var_format is not None:
        var_format = var_format.replace('{:', '').replace('}', '')  # no {}

    sns.heatmap(data=df,
                center=0,
                annot=True,
                fmt=var_format,
                cmap=cmap,
                alpha=alpha,
                cbar_kws={'size': fontsize},
                cbar=False,
                annot_kws={'size': fontsize},
                xticklabels=True,  # important for full display of labels
                yticklabels=True,  # important for full display of labels
                square=square,
                vmin=vmin,
                vmax=vmax,
                ax=ax)  # ,"ha": 'right' #cbar_kws={'format': '%0.2f%%'}

    if top_x_label:
        ax.xaxis.tick_top()

    if not transpose:
        pass
        # bottom, top = ax.get_ylim()
        # ax.set_ylim(bottom + 0.5, top - 0.5)
    else:
        ax.xaxis.labelpad = labelpad

    put.set_ax_tick_params(ax=ax, fontsize=fontsize, labelbottom=not top_x_label, labeltop=top_x_label, **kwargs)
    put.set_ax_tick_labels(ax=ax, fontsize=fontsize, **kwargs)

    if vline_columns is not None:
        for vline_column in vline_columns:
            ax.vlines([vline_column], *ax.get_ylim(), lw=1)

    if hline_rows is not None:
        for hline_row in hline_rows:
            ax.hlines([hline_row], *ax.get_xlim(), lw=1)

    if title is not None:
        put.set_title(ax=ax, title=title, fontsize=fontsize, **kwargs)

    ax.set_ylabel(ylabel)
    ax.set_xlabel('')

    return fig


class UnitTests(Enum):
    HEATMAP = 1


def run_unit_test(unit_test: UnitTests):

    from qis.test_data import load_etf_data
    prices = load_etf_data().dropna()

    if unit_test == UnitTests.HEATMAP:
        corrs = prices.pct_change().corr()
        plot_heatmap(corrs, inverse=False, x_rotation=90)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.HEATMAP

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)
